import asyncio
from collections import Counter

from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template

few_shot_examples = """
Q: There are 15 trees in the grove. Grove workers will plant trees in the grove today. After they are done,
there will be 21 trees. How many trees did the grove workers plant today?
A: We start with 15 trees. Later we have 21 trees. The difference must be the number of trees they planted.
So, they must have planted 21 - 15 = 6 trees. The answer is 6.
Q: If there are 3 cars in the parking lot and 2 more cars arrive, how many cars are in the parking lot?
A: There are 3 cars in the parking lot already. 2 more arrive. Now there are 3 + 2 = 5 cars. The answer is 5.
Q: Leah had 32 chocolates and her sister had 42. If they ate 35, how many pieces do they have left in total?
A: Leah had 32 chocolates and Leah’s sister had 42. That means there were originally 32 + 42 = 74
chocolates. 35 have been eaten. So in total they still have 74 - 35 = 39 chocolates. The answer is 39.
Q: Jason had 20 lollipops. He gave Denny some lollipops. Now Jason has 12 lollipops. How many lollipops
did Jason give to Denny?
A: Jason had 20 lollipops. Since he only has 12 now, he must have given the rest to Denny. The number of
lollipops he has given to Denny must have been 20 - 12 = 8 lollipops. The answer is 8.
Q: Shawn has five toys. For Christmas, he got two toys each from his mom and dad. How many toys does
he have now?
A: He has 5 toys. He got 2 from mom, so after that he has 5 + 2 = 7 toys. Then he got 2 more from dad, so
in total he has 7 + 2 = 9 toys. The answer is 9.
Q: There were nine computers in the server room. Five more computers were installed each day, from
monday to thursday. How many computers are now in the server room?
A: There are 4 days from monday to thursday. 5 computers were added each day. That means in total 4 * 5 =
20 computers were added. There were 9 computers in the beginning, so now there are 9 + 20 = 29 computers.
The answer is 29.
Q: Michael had 58 golf balls. On tuesday, he lost 23 golf balls. On wednesday, he lost 2 more. How many
golf balls did he have at the end of wednesday?
A: Michael initially had 58 balls. He lost 23 on Tuesday, so after that he has 58 - 23 = 35 balls. On
Wednesday he lost 2 more so now he has 35 - 2 = 33 balls. The answer is 33.
"""


@openai.call(model="gpt-4o-mini", call_params={"temperature": 0.5})
@prompt_template(
    """
    Some examples on how to think step by step:
    {few_shot_examples}

    Answer the following question, thinking step by step:
    {query}
    """
)
async def chain_of_thought(query: str) -> openai.OpenAIDynamicConfig:
    return {"computed_fields": {"few_shot_examples": few_shot_examples}}


class Solution(BaseModel):
    solution_value: int = Field(
        ..., description="The actual number of a solution to a math problem."
    )


@openai.call(model="gpt-4o-mini", response_model=Solution)
@prompt_template(
    """
    Extract just the number of a solution to a math problem.
    For example, for the solution:
    Michael initially had 58 balls. He lost 23 on Tuesday, so after that he has
    58 - 23 = 35 balls. On Wednesday he lost 2 more so now he has 35 - 2 = 33 balls.
    The answer is 33.
    
    You would extract 33.

    Solution to extract from:
    {response}
    """
)
async def extract_number(response: str): ...


def most_frequent(lst):
    """Returns the most frequent element in a list."""
    counter = Counter(lst)
    most_common = counter.most_common(1)
    return most_common[0][0] if most_common else None


async def self_consistency(query: str, num_samples: int):
    cot_tasks = [chain_of_thought(query) for _ in range(num_samples)]
    cot_responses = [response.content for response in await asyncio.gather(*cot_tasks)]
    # Uncomment to see intermediate responses
    # for response in cot_responses:
    #     print(response)
    extract_number_tasks = [extract_number(response) for response in cot_responses]
    response_numbers = [
        response.solution_value
        for response in await asyncio.gather(*extract_number_tasks)
    ]
    return most_frequent(response_numbers)


query = """Olivia has $23. She bought five bagels for $3 each.
How much money does she have left?"""

print(asyncio.run(self_consistency(query=query, num_samples=5)))
# > 8


# Intermediate Responses

# cot_responses
# To find out how much money Olivia has left after buying the bagels, we can follow these steps:

# 1. **Determine the total cost of the bagels**: Olivia bought 5 bagels, and each bagel costs $3.
#    - Total cost = Number of bagels × Cost per bagel
#    - Total cost = 5 × 3 = $15

# 2. **Subtract the total cost from Olivia's initial amount**: Olivia started with $23, and we need to subtract the total cost of the bagels from this amount.
#    - Money left = Initial amount - Total cost
#    - Money left = 23 - 15 = $8

# Therefore, Olivia has $8 left after buying the bagels. The answer is $8.
# To find out how much money Olivia has left after buying the bagels, we can follow these steps:

# 1. **Determine the cost of the bagels**: Olivia bought 5 bagels, and each bagel costs $3. So, we calculate the total cost:
#    \[
#    \text{Total cost} = 5 \times 3 = 15 \text{ dollars}
#    \]

# 2. **Calculate how much money Olivia has left**: Olivia started with $23. After spending $15 on the bagels, we subtract the cost from her initial amount:
#    \[
#    \text{Money left} = 23 - 15 = 8 \text{ dollars}
#    \]

# Therefore, Olivia has $8 left. The answer is 8.
# To find out how much money Olivia has left, we can follow these steps:

# 1. **Determine the cost of the bagels**: Olivia bought 5 bagels, and each bagel costs $3. So, we calculate the total cost:
#    \[
#    5 \text{ bagels} \times 3 \text{ dollars/bagel} = 15 \text{ dollars}
#    \]

# 2. **Subtract the cost from her initial amount**: Olivia started with $23. Now we subtract the total cost of the bagels from this amount:
#    \[
#    23 \text{ dollars} - 15 \text{ dollars} = 8 \text{ dollars}
#    \]

# 3. **Conclusion**: After buying the bagels, Olivia has $8 left.

# The answer is $8.
# To find out how much money Olivia has left after buying the bagels, let's go through the steps:

# 1. **Determine how much Olivia spent on the bagels.**
#    Olivia bought 5 bagels for $3 each.
#    To find the total cost of the bagels, we multiply the number of bagels by the price per bagel:
#    \[
#    5 \text{ bagels} \times 3 \text{ dollars/bagel} = 15 \text{ dollars}
#    \]

# 2. **Calculate how much money Olivia has left after the purchase.**
#    Olivia started with $23 and spent $15 on the bagels.
#    To find out how much money she has left, we subtract the amount spent from her initial amount:
#    \[
#    23 \text{ dollars} - 15 \text{ dollars} = 8 \text{ dollars}
#    \]

# Therefore, Olivia has $8 left. The answer is $8.
# To find out how much money Olivia has left after buying the bagels, we can follow these steps:

# 1. **Determine the cost of the bagels**: Olivia bought 5 bagels for $3 each.
#    - So, the total cost for the bagels is:
#    \[
#    5 \text{ bagels} \times 3 \text{ dollars/bagel} = 15 \text{ dollars}
#    \]

# 2. **Calculate how much money Olivia has left**: Olivia started with $23, and now we subtract the cost of the bagels from her initial amount.
#    - The calculation is:
#    \[
#    23 \text{ dollars} - 15 \text{ dollars} = 8 \text{ dollars}
#    \]

# 3. **Conclusion**: After buying the bagels, Olivia has $8 left.

# The answer is $8.



"""This module contains the base classes for streaming responses from LLMs."""

import datetime
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator
from functools import wraps
from typing import (
    Any,
    ClassVar,
    Generic,
    ParamSpec,
    TypeVar,
    overload,
)

from ._utils import (
    HandleStream,
    HandleStreamAsync,
    SetupCall,
    get_fn_args,
    get_metadata,
    get_possible_user_message_param,
)
from ._utils._protocols import fn_is_async
from .call_kwargs import BaseCallKwargs
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .metadata import Metadata
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_UserMessageParamT = TypeVar("_UserMessageParamT")
_AssistantMessageParamT = TypeVar("_AssistantMessageParamT")
_ToolMessageParamT = TypeVar("_ToolMessageParamT")
_MessageParamT = TypeVar("_MessageParamT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ToolSchemaT = TypeVar("_ToolSchemaT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_FinishReason = TypeVar("_FinishReason")
_DEFAULT = object()


class BaseStream(
    Generic[
        _BaseCallResponseT,
        _BaseCallResponseChunkT,
        _UserMessageParamT,
        _AssistantMessageParamT,
        _ToolMessageParamT,
        _MessageParamT,
        _BaseToolT,
        _ToolSchemaT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _FinishReason,
    ],
    ABC,
):
    """A base class for streaming responses from LLMs."""

    stream: (
        Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]
        | AsyncGenerator[
            tuple[_BaseCallResponseChunkT, _BaseToolT | None],
            None,
        ]
    )
    content: str
    metadata: Metadata
    tool_types: list[type[_BaseToolT]] | None
    call_response_type: type[_BaseCallResponseT]
    model: str
    prompt_template: str | None = None
    fn_args: dict[str, Any] | None = None
    dynamic_config: _BaseDynamicConfigT
    messages: list[_MessageParamT]
    call_params: _BaseCallParamsT
    call_kwargs: BaseCallKwargs[_ToolSchemaT]
    user_message_param: _UserMessageParamT | None = None
    message_param: _AssistantMessageParamT
    input_tokens: int | float | None = None
    output_tokens: int | float | None = None
    id: str | None = None
    finish_reasons: list[_FinishReason] | None = None
    start_time: float = 0
    end_time: float = 0

    _provider: ClassVar[str] = "NO PROVIDER"

    def __init__(
        self,
        *,
        stream: Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]
        | AsyncGenerator[
            tuple[_BaseCallResponseChunkT, _BaseToolT | None],
            None,
        ],
        metadata: Metadata,
        tool_types: list[type[_BaseToolT]] | None,
        call_response_type: type[_BaseCallResponseT],
        model: str,
        prompt_template: str | None,
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        messages: list[_MessageParamT],
        call_params: _BaseCallParamsT,
        call_kwargs: BaseCallKwargs[_ToolSchemaT],
    ) -> None:
        """Initializes an instance of `BaseStream`."""
        self.content = ""
        self.stream = stream
        self.metadata = metadata
        self.tool_types = tool_types
        self.call_response_type = call_response_type
        self.model = model
        self.prompt_template = prompt_template
        self.fn_args = fn_args
        self.dynamic_config = dynamic_config
        self.messages = messages
        self.call_params = call_params
        self.call_kwargs = call_kwargs
        self.user_message_param = get_possible_user_message_param(messages)  # pyright: ignore [reportAttributeAccessIssue]

    def __iter__(
        self,
    ) -> Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]:
        """Iterator over the stream and stores useful information."""
        assert isinstance(
            self.stream, Generator
        ), "Stream must be a generator for __iter__"
        self.content, tool_calls = "", []
        self.start_time = datetime.datetime.now().timestamp() * 1000
        for chunk, tool in self.stream:
            self._update_properties(chunk)
            if tool:
                tool_call = getattr(tool, "tool_call", _DEFAULT)
                if tool_call != _DEFAULT:
                    tool_calls.append(tool_call)
            yield chunk, tool
        self.end_time = datetime.datetime.now().timestamp() * 1000
        self.message_param = self._construct_message_param(
            tool_calls or None, self.content
        )

    def __aiter__(
        self,
    ) -> AsyncGenerator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None]:
        """Iterates over the stream and stores useful information."""
        self.content = ""

        async def generator() -> (
            AsyncGenerator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None]
        ):
            assert isinstance(
                self.stream, AsyncGenerator
            ), "Stream must be an async generator for __aiter__"
            tool_calls = []
            async for chunk, tool in self.stream:
                self._update_properties(chunk)
                if tool:
                    tool_call = getattr(tool, "tool_call", _DEFAULT)
                    if tool_call != _DEFAULT:
                        tool_calls.append(tool_call)
                yield chunk, tool
            self.message_param = self._construct_message_param(
                tool_calls or None, self.content
            )

        return generator()

    def _update_properties(self, chunk: _BaseCallResponseChunkT) -> None:
        """Updates the properties of the stream."""
        self.content += chunk.content
        if chunk.input_tokens is not None:
            self.input_tokens = (
                chunk.input_tokens
                if not self.input_tokens
                else self.input_tokens + chunk.input_tokens
            )
        if chunk.output_tokens is not None:
            self.output_tokens = (
                chunk.output_tokens
                if not self.output_tokens
                else self.output_tokens + chunk.output_tokens
            )
        if chunk.model is not None:
            self.model = chunk.model
        if chunk.id is not None:
            self.id = chunk.id
        if chunk.finish_reasons is not None:
            self.finish_reasons = chunk.finish_reasons

    @property
    @abstractmethod
    def cost(self) -> float | None:
        """Returns the cost of the stream."""
        ...

    @abstractmethod
    def _construct_message_param(
        self, tool_calls: list[Any] | None = None, content: str | None = None
    ) -> _AssistantMessageParamT:
        """Constructs the assistant message."""
        ...

    def tool_message_params(
        self, tools_and_outputs: list[tuple[_BaseToolT, str]]
    ) -> list[_ToolMessageParamT]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The list of tools and their outputs from which the tool
                message parameters should be constructed.
        """
        return self.call_response_type.tool_message_params(tools_and_outputs)

    @abstractmethod
    def construct_call_response(self) -> _BaseCallResponseT:
        """Constructs the call response."""
        ...


_BaseClientT = TypeVar("_BaseClientT", bound=object)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_P = ParamSpec("_P")


def stream_factory(  # noqa: ANN201
    *,
    TCallResponse: type[_BaseCallResponseT],
    TStream: type[BaseStream],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
    handle_stream: HandleStream[_ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT],
    handle_stream_async: HandleStreamAsync[
        _ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT
    ],
):
    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, TStream]: ...

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, Awaitable[TStream]]: ...

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT]
        | Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, TStream] | Callable[_P, Awaitable[TStream]]:
        if fn_is_async(fn):

            @wraps(fn)
            async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> TStream:
                fn_args = get_fn_args(fn, args, kwargs)
                dynamic_config = await fn(*args, **kwargs)
                create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                    model=model,
                    client=client,
                    fn=fn,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    tools=tools,
                    json_mode=json_mode,
                    call_params=call_params,
                    extract=False,
                )

                async def generator() -> (
                    AsyncGenerator[
                        tuple[_BaseCallResponseChunkT, _BaseToolT | None], None
                    ]
                ):
                    async for chunk, tool in handle_stream_async(
                        await create(stream=True, **call_kwargs), tool_types
                    ):
                        yield chunk, tool

                return TStream(
                    stream=generator(),
                    metadata=get_metadata(fn, dynamic_config),
                    tool_types=tool_types,  # pyright: ignore [reportArgumentType]
                    call_response_type=TCallResponse,
                    model=model,
                    prompt_template=prompt_template,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    messages=messages,
                    call_params=call_params,
                    call_kwargs=call_kwargs,
                )

            return inner_async
        else:

            @wraps(fn)
            def inner(*args: _P.args, **kwargs: _P.kwargs) -> TStream:
                fn_args = get_fn_args(fn, args, kwargs)
                dynamic_config = fn(*args, **kwargs)
                create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                    model=model,
                    client=client,
                    fn=fn,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    tools=tools,
                    json_mode=json_mode,
                    call_params=call_params,
                    extract=False,
                )

                def generator() -> (
                    Generator[
                        tuple[_BaseCallResponseChunkT, _BaseToolT | None],
                        None,
                        None,
                    ]
                ):
                    yield from handle_stream(
                        create(stream=True, **call_kwargs), tool_types
                    )

                return TStream(
                    stream=generator(),
                    metadata=get_metadata(fn, dynamic_config),
                    tool_types=tool_types,  # pyright: ignore [reportArgumentType]
                    call_response_type=TCallResponse,
                    model=model,
                    prompt_template=prompt_template,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    messages=messages,
                    call_params=call_params,
                    call_kwargs=call_kwargs,
                )

            return inner

    return decorator



"""The `AzureStream` class for convenience around streaming LLM calls.

usage docs: learn/streams.md
"""

import datetime

from azure.ai.inference.models import (
    AssistantMessage,
    ChatChoice,
    ChatCompletions,
    ChatCompletionsToolCall,
    ChatCompletionsToolDefinition,
    ChatRequestMessage,
    ChatResponseMessage,
    CompletionsFinishReason,
    CompletionsUsage,
    FunctionCall,
    ToolMessage,
    UserMessage,
)

from ..base.stream import BaseStream
from ._utils import calculate_cost
from .call_params import AzureCallParams
from .call_response import AzureCallResponse
from .call_response_chunk import AzureCallResponseChunk
from .dynamic_config import AzureDynamicConfig
from .tool import AzureTool


class AzureStream(
    BaseStream[
        AzureCallResponse,
        AzureCallResponseChunk,
        UserMessage,
        AssistantMessage,
        ToolMessage,
        ChatRequestMessage,
        AzureTool,
        ChatCompletionsToolDefinition,
        AzureDynamicConfig,
        AzureCallParams,
        CompletionsFinishReason,
    ]
):
    """A class for convenience around streaming Azure LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.azure import azure_call


    @azure_call("gpt-4o-mini", stream=True)
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    stream = recommend_book("fantasy")  # returns `AzureStream` instance
    for chunk, _ in stream:
        print(chunk.content, end="", flush=True)
    ```
    """

    _provider = "azure"

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    def _construct_message_param(
        self,
        tool_calls: list[ChatCompletionsToolCall] | None = None,
        content: str | None = None,
    ) -> AssistantMessage:
        """Constructs the message parameter for the assistant."""
        message_param = AssistantMessage(content=content)
        if tool_calls:
            message_param["tool_calls"] = [
                ChatCompletionsToolCall(
                    function=FunctionCall(
                        arguments=tool_call.function.arguments,
                        name=tool_call.function.name,
                    ),
                    id=tool_call.id,
                )
                for tool_call in tool_calls
            ]
        return message_param

    def construct_call_response(self) -> AzureCallResponse:
        """Constructs the call response from a consumed AzureStream.

        Raises:
            ValueError: if the stream has not yet been consumed.
        """
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        message = ChatResponseMessage(
            role=self.message_param["role"],
            content=self.message_param.get("content", ""),
            tool_calls=self.message_param.get("tool_calls", []),
        )
        if not self.input_tokens and not self.output_tokens:
            usage = CompletionsUsage(
                completion_tokens=0, prompt_tokens=0, total_tokens=0
            )
        else:
            usage = CompletionsUsage(
                prompt_tokens=int(self.input_tokens or 0),
                completion_tokens=int(self.output_tokens or 0),
                total_tokens=int(self.input_tokens or 0) + int(self.output_tokens or 0),
            )
        completion = ChatCompletions(
            id=self.id if self.id else "",
            model=self.model,
            choices=[
                ChatChoice(
                    finish_reason=self.finish_reasons[0]
                    if self.finish_reasons
                    else "stop",
                    index=0,
                    message=message,
                )
            ],
            created=datetime.datetime.now(),
            usage=usage,
        )
        return AzureCallResponse(
            metadata=self.metadata,
            response=completion,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )



import base64

import httpx
from pydantic import BaseModel, Field

from mirascope.core import anthropic, openai, prompt_template

image_url = "https://www.receiptfont.com/wp-content/uploads/template-mcdonalds-1-screenshot-fit.png"

image_media_type = "image/png"
image_data = base64.b64encode(httpx.get(image_url).content).decode("utf-8")


class Item(BaseModel):
    name: str = Field(..., description="The name of the item")
    quantity: int = Field(..., description="The quantity of the item")
    price: float = Field(..., description="The price of the item")


@openai.call(model="gpt-4o", response_model=list[Item])
@prompt_template(
    """
    SYSTEM:
    Extract the receipt information from the image.
    
    USER:
    {url:image}
    """
)
def extract_receipt_info_openai(url: str): ...


@anthropic.call(
    model="claude-3-5-sonnet-20240620", response_model=list[Item], json_mode=True
)
@prompt_template(
    """
    Extract the receipt information from the image.
    
    {url:image}
    """
)
def extract_receipt_info_anthropic(url: str): ...


print(extract_receipt_info_openai(image_url))
print(extract_receipt_info_anthropic(image_url))



"""Tests the `_utils.json_mode_content` module."""

from mirascope.core.base._utils._json_mode_content import json_mode_content
from mirascope.core.base.tool import BaseTool


def test_json_mode_content() -> None:
    """Tests the `json_mode_content` function."""

    class Book(BaseTool):
        title: str

    assert (
        json_mode_content(None)
        == "\n\nExtract ONLY a valid JSON dict using the schema."
    )
    assert (
        json_mode_content(Book)
        == """

Extract ONLY a valid JSON dict (NOT THE SCHEMA) from the content that adheres to this schema:
{
  "properties": {
    "title": {
      "type": "string"
    }
  },
  "required": [
    "title"
  ],
  "type": "object"
}"""
    )



from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel

from mirascope.core import openai, prompt_template


class Librarian(BaseModel):
    history: list[ChatCompletionMessageParam] = []

    @openai.call(model="gpt-4o-mini")
    @prompt_template(
        """
        SYSTEM: You are a helpful librarian assistant.
        MESSAGES: {self.history}
        USER: {query}
        """
    )
    def _call(self, query: str) -> openai.OpenAIDynamicConfig: ...

    def run(self):
        while True:
            query = input("User: ")
            if query.lower() == "exit":
                break

            response = self._call(query)
            print(f"Librarian: {response.content}")

            if response.user_message_param:
                self.history.append(response.user_message_param)
            self.history.append(response.message_param)


librarian = Librarian()
librarian.run()



from typing import Any, Literal

import pandas as pd
from dotenv import load_dotenv
from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template

load_dotenv()


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Generate {num_datapoints} random but realistic datapoints of items which could
    be in a home appliances warehouse. Output the datapoints as a csv, and your
    response should only be the CSV.

    Format:
    Name, Price, Inventory

    Name - the name of the home appliance product
    Price - the price of an individual product, in dollars (include cents)
    Inventory - how many are left in stock
    """
)
def generate_csv_data(num_datapoints: int): ...


print(generate_csv_data(5))


class HomeAppliance(BaseModel):
    name: str = Field(description="The name of the home appliance product")
    price: float = Field(
        description="The price of an individual product, in dollars (include cents)"
    )
    inventory: int = Field(description="How many of the items are left in stock")


@openai.call(model="gpt-4o-mini", response_model=list[HomeAppliance])
@prompt_template(
    """
    Generate {num_datapoints} random but realistic datapoints of items which could
    be in a home appliances warehouse. Output the datapoints as a list of instances
    of HomeAppliance.
    """
)
def generate_home_appliance_data(num_datapoints: int): ...


print(generate_home_appliance_data(5))


class DataFrameGenerator(BaseModel):
    data: list[list[Any]] = Field(
        description="the data to be inserted into the dataframe"
    )
    column_names: list[str] = Field(description="The names of the columns in data")

    def append_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        return pd.concat([df, self.generate_dataframe()], ignore_index=True)

    def generate_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(dict(zip(self.column_names, self.data, strict=False)))


@openai.call(model="gpt-4o-mini", response_model=DataFrameGenerator)
@prompt_template(
    """
    Generate {num_datapoints} random but realistic datapoints of items which could
    be in a home appliances warehouse. Generate your response as `data` and
    `column_names`, so that a pandas DataFrame may be generated with:
    `pd.DataFrame(data, columns=column_names)`.

    Format:
    Name, Price, Inventory

    Name - the name of the home appliance product
    Price - the price of an individual product, in dollars (include cents)
    Inventory - how many are left in stock
    """
)
def generate_df_data(num_datapoints: int): ...


df_data = generate_df_data(5)
df = df_data.generate_dataframe()


@openai.call(model="gpt-4o-mini", response_model=DataFrameGenerator)
@prompt_template(
    """
    Generate {num_datapoints} random but realistic datapoints of items which would
    make sense to the following dataset:
    {df}
    Generate your response as `data` and
    `column_names`, so that a pandas DataFrame may be generated with:
    `pd.DataFrame(data, columns=column_names)` then appended to the existing data.
    """
)
def generate_additional_df_data(num_datapoints: int, df: pd.DataFrame): ...


df_data = generate_additional_df_data(5, df)
df = df_data.append_dataframe(df)


class TV(BaseModel):
    size: int = Field(description="The size of the TV")
    price: float = Field(description="The price of the TV in dollars (include cents)")
    tv_type: Literal["OLED", "QLED"]


@openai.call(model="gpt-4o-mini", response_model=list[TV])
@prompt_template(
    """
    Generate {num_datapoints} random but realistic datapoints of TVs.
    Output the datapoints as a list of instances of TV.

    Make sure to abide by the following constraints:
    QLEDS should be roughly (not exactly) 2x the price of an OLED of the same size
    for both OLEDs and QLEDS, price should increase roughly proportionately to size
    """
)
def generate_tv_data(num_datapoints: int): ...


for tv in generate_tv_data(10):
    print(tv)



"""This module contains the config dict for response models."""

from pydantic import ConfigDict


class ResponseModelConfigDict(ConfigDict, total=False):
    """An extension of the base `ConfigDict` for configuring response models."""

    strict: bool



import asyncio
import sqlite3
from typing import ClassVar
from unittest.mock import MagicMock

import pytest
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from pydantic import BaseModel, Field

from examples.cookbook.agents.sql_agent.agent import Librarian
from mirascope.core import (
    BasePrompt,
    anthropic,
    openai,
    prompt_template,
)


@pytest.fixture
def mock_librarian():
    class MockLibrarian(Librarian):
        con: ClassVar[sqlite3.Connection] = MagicMock()

    return MockLibrarian()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "select_query",
    [
        "Get all books",
        "Get every single book",
        "Show me all books",
        "List all books",
        "Display all books",
    ],
)
async def test_select_query(select_query: str, mock_librarian: Librarian):
    response = await mock_librarian._stream(select_query)
    async for _, tool in response:
        query = tool.args.get("query", "") if tool else ""
        assert query == "SELECT * FROM ReadingList;"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "insert_query",
    [
        "Please add Gone with the Wind to my reading list",
        "You recently recommended Gone with the Wind, can you add it to my reading list.",
    ],
)
async def test_insert_query(insert_query: str, mock_librarian: Librarian):
    response = await mock_librarian._stream(insert_query)
    async for _, tool in response:
        query = tool.args.get("query", "") if tool else ""
        assert (
            query
            == "INSERT INTO ReadingList (title, status) VALUES ('Gone with the Wind', 'Not Started');"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "update_query",
    [
        "Can you mark Gone with the Wind as read?",
        "I just finished Gone with the Wind, can you update the status?",
    ],
)
async def test_update_query(update_query: str, mock_librarian: Librarian):
    response = await mock_librarian._stream(update_query)
    async for _, tool in response:
        query = tool.args.get("query", "") if tool else ""
        assert (
            query
            == "UPDATE ReadingList SET status = 'Complete' WHERE title = 'Gone with the Wind';"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "delete_query",
    [
        "Can you remove Gone with the Wind from my reading list?",
        "Can you delete Gone with the Wind?",
    ],
)
async def test_delete_query(delete_query: str, mock_librarian: Librarian):
    response = await mock_librarian._stream(delete_query)
    async for _, tool in response:
        query = tool.args.get("query", "") if tool else ""
        assert query == "DELETE FROM ReadingList WHERE title = 'Gone with the Wind';"


@pytest.mark.asyncio
async def test_conversation(mock_librarian: Librarian):
    mock_librarian.messages = [
        ChatCompletionUserMessageParam(
            role="user", content="Can you recommend me a fantasy book?"
        ),
        ChatCompletionAssistantMessageParam(
            role="assistant",
            content="I would recommend 'The Name of the Wind' by Patrick Rothfuss. It’s the first book in 'The Kingkiller Chronicle' series and features a beautifully written story about a gifted young man who grows up to be a legendary figure. It's filled with magic, adventure, and rich character development. I believe you'll enjoy it! Would you like to add it to your reading list?",
        ),
    ]
    response = await mock_librarian._stream("Can you add it to my reading list?")
    async for _, tool in response:
        query = tool.args.get("query", "") if tool else ""
        assert (
            query
            == "INSERT INTO ReadingList (title, status) VALUES ('The Name of the Wind', 'Not Started');"
        )


@prompt_template(
    """
    Evaluate the friendliness of generated text on a continuous scale from 1 to 5.
    A generation is friendly (score > 1) if it contains content that is:

    - Warm, welcoming, and positive in tone
    - Respectful, considerate, and empathetic towards others
    - Encouraging, supportive, or promotes kindness
    - Inclusive and mindful of diverse perspectives

    Provide a brief reasoning for your assigned score.
    Make sure to highlight the specific aspects that impacted your score.

    Use the following friendliness scale:

    1 - No friendliness; cold, distant, or neutral tone
    2 - Mild friendliness; slightly warm or polite language
    3 - Moderate friendliness; noticeably kind and welcoming content
    4 - High friendliness; very warm, supportive, and empathetic language
    5 - Extreme friendliness; exceptionally kind, caring, and uplifting content

    Input Query: {input_query}
    Output Content: {output_content}
    """
)
class FriendlinessEvaluator(BasePrompt):
    input_query: str
    output_content: str


@pytest.mark.asyncio
async def test_friendliness(mock_librarian: Librarian):
    input_query = "Can you add gone with the wind"
    mock_librarian.messages = [
        ChatCompletionUserMessageParam(role="user", content=input_query),
        ChatCompletionAssistantMessageParam(
            role="assistant",
            tool_calls=[
                {
                    "type": "function",
                    "function": {
                        "arguments": "{\"query\":\"INSERT INTO ReadingList (title, status) VALUES ('Gone with the Wind', 'Not Started');\"}",
                        "name": "_execute_query",
                    },
                    "id": "1",
                }
            ],
        ),
        ChatCompletionToolMessageParam(
            role="tool",
            content="Query executed successfully, 1 row(s) were updated/inserted.",
            tool_call_id="1",
        ),
    ]
    response = await mock_librarian._stream("")
    output_content = ""
    async for chunk, _ in response:
        output_content += chunk.content
    prompt = FriendlinessEvaluator(
        input_query=input_query, output_content=output_content
    )

    class Eval(BaseModel):
        score: float = Field(..., description="A score between [1.0, 5.0]")
        reasoning: str = Field(..., description="The reasoning for the score")

    async def run_evals() -> list[Eval]:
        judges = [
            openai.call(
                "gpt-4o-mini",
                response_model=Eval,
                json_mode=True,
            ),
            anthropic.call(
                "claude-3-5-sonnet-20240620",
                response_model=Eval,
                json_mode=True,
            ),
        ]
        calls = [prompt.run_async(judge) for judge in judges]
        return await asyncio.gather(*calls)

    evals = await run_evals()
    total_score = sum(eval.score for eval in evals)
    average_score = total_score / len(evals)
    assert average_score > 2



from fastapi import FastAPI
from pydantic import BaseModel

from mirascope.core import openai, prompt_template

app = FastAPI()


class Book(BaseModel):
    title: str
    author: str


@app.post("/recommend_book")
@openai.call(model="gpt-4o-mini", response_model=Book)
@prompt_template("Recommend a {genre} book.")
def recommend_book(genre: str): ...


@app.post("/recommend_book_async")
@openai.call(model="gpt-4o-mini", response_model=Book)
@prompt_template("Recommend a {genre} book.")
async def recommend_book_async(genre: str): ...



from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel

from mirascope.core.base import BaseCallResponse
from mirascope.core.base.stream import BaseStream
from mirascope.core.base.structured_stream import BaseStructuredStream
from mirascope.integrations._middleware_factory import (
    default_context_manager,
    middleware_factory,
)


def test_default_context_manager() -> None:
    with default_context_manager(lambda x: x) as result:
        assert result is None


def test_default_context_manager_async() -> None:
    async def async_fn(x):
        return x  # pragma: no cover

    with default_context_manager(async_fn) as result:
        assert result is None


def test_middleware_factory_call_response_sync() -> None:
    class MyCallResponse(BaseCallResponse):
        @property
        def content(self) -> str:
            return "content"

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )  # type: ignore

    def sync_fn() -> MyCallResponse:
        return call_response

    def handle_call_response(result, fn, context) -> None:
        assert isinstance(result, BaseCallResponse)

    decorate = middleware_factory(handle_call_response=handle_call_response)(sync_fn)
    result = decorate()
    assert result.content == call_response.content


@pytest.mark.asyncio
async def test_middleware_factory_call_response_async() -> None:
    class MyCallResponse(BaseCallResponse):
        @property
        def content(self) -> str:
            return "content"

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )  # type: ignore

    async def async_fn() -> MyCallResponse:
        return call_response

    async def handle_call_response_async(result, fn, context) -> None:
        assert isinstance(result, BaseCallResponse)

    decorate = middleware_factory(
        handle_call_response_async=handle_call_response_async
    )(async_fn)
    result = await decorate()
    assert result.content == call_response.content


def test_middleware_factory_stream_sync() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    mock_chunk = MagicMock()
    mock_chunk.content = "content"
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    class MyStream(BaseStream):
        @property
        def cost(self) -> int:
            return 10

    # Usage
    stream_chunks = [(mock_chunk, None)]
    my_stream = MyStream(
        stream=(t for t in stream_chunks),
        metadata={},
        tool_types=[],
        call_response_type=MagicMock,
        model="model",
        prompt_template="prompt_template",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
    )  # type: ignore

    def sync_fn() -> BaseStream:
        return my_stream

    def handle_stream(result, fn, context) -> None:
        assert isinstance(result, BaseStream)

    decorate = middleware_factory(handle_stream=handle_stream)(sync_fn)
    result = decorate()
    result_chunks = []
    for chunk, _ in result:
        result_chunks.append(chunk.content)
    for result_chunk, stream_chunk in zip(result_chunks, stream_chunks, strict=False):
        assert result_chunk == stream_chunk[0].content
    assert my_stream.cost == result.cost


@pytest.mark.asyncio
async def test_middleware_factory_stream_async() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    mock_chunk = MagicMock()
    mock_chunk.content = "content"
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    class MyStream(BaseStream):
        @property
        def cost(self) -> int:
            return 10

    # Usage
    stream_chunks = [(mock_chunk, None)]
    my_stream = MyStream(
        stream=(t for t in stream_chunks),
        metadata={},
        tool_types=[],
        call_response_type=MagicMock,
        model="model",
        prompt_template="prompt_template",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
    )  # type: ignore

    async def generator():
        yield mock_chunk, None

    my_stream.stream = generator()

    async def async_fn() -> BaseStream:
        return my_stream

    async def handle_stream_async(result, fn, context) -> None:
        assert isinstance(result, BaseStream)

    decorate = middleware_factory(handle_stream_async=handle_stream_async)(async_fn)
    result = await decorate()
    result_chunks = []
    async for chunk, _ in result:
        result_chunks.append(chunk.content)
    for result_chunk, stream_chunk in zip(result_chunks, stream_chunks, strict=False):
        assert result_chunk == stream_chunk[0].content
    assert my_stream.cost == result.cost


def test_middleware_factory_base_model_sync() -> None:
    class Foo(BaseModel):
        bar: str
        baz: int

    def sync_fn() -> Foo:
        return Foo(bar="bar", baz=1)

    def handle_response_model(result, fn, context) -> None:
        assert isinstance(result, BaseModel)

    decorate = middleware_factory(handle_response_model=handle_response_model)(sync_fn)
    result = decorate()
    assert result.model_dump() == sync_fn().model_dump()


@pytest.mark.asyncio
async def test_middleware_factory_base_model_async() -> None:
    class Foo(BaseModel):
        bar: str
        baz: int

    async def async_fn() -> Foo:
        return Foo(bar="bar", baz=1)

    async def handle_response_model_async(result, fn, context) -> None:
        assert isinstance(result, BaseModel)

    decorate = middleware_factory(
        handle_response_model_async=handle_response_model_async
    )(async_fn)
    result = await decorate()
    async_fn_result = await async_fn()
    assert result.model_dump() == async_fn_result.model_dump()


def test_middleware_factory_structured_stream() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class Foo(BaseModel):
        bar: str
        baz: int

    mock_chunk = MagicMock()
    my_foo = Foo(bar="bar", baz=1)
    mock_chunk.content = my_foo.model_dump_json()
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    base_stream = MagicMock()
    base_stream.__iter__.return_value = (t for t in [(mock_chunk, None)])
    my_structured_stream = BaseStructuredStream(stream=base_stream, response_model=Foo)

    def sync_fn() -> BaseStructuredStream:
        return my_structured_stream

    def handle_structured_stream(result, fn, context) -> None:
        assert isinstance(result, BaseStructuredStream)

    decorate = middleware_factory(handle_structured_stream=handle_structured_stream)(
        sync_fn
    )
    for chunk in decorate():
        assert chunk.model_dump() == my_foo.model_dump()


@pytest.mark.asyncio
async def test_middleware_factory_structured_stream_async() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class Foo(BaseModel):
        bar: str
        baz: int

    mock_chunk = MagicMock()
    my_foo = Foo(bar="bar", baz=1)
    mock_chunk.content = my_foo.model_dump_json()
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    base_stream = MagicMock()

    async def generator(self):
        yield mock_chunk, None

    base_stream.__aiter__ = generator
    my_structured_stream = BaseStructuredStream(stream=base_stream, response_model=Foo)

    async def async_fn() -> BaseStructuredStream:
        return my_structured_stream

    async def handle_structured_stream_async(result, fn, context) -> None:
        assert isinstance(result, BaseStructuredStream)

    decorate = middleware_factory(
        handle_structured_stream_async=handle_structured_stream_async
    )(async_fn)
    async for chunk in await decorate():
        assert chunk.model_dump() == my_foo.model_dump()



"""This module contains the type definition for the base call keyword arguments."""

from typing import Generic, TypeVar

from typing_extensions import NotRequired

from .call_params import BaseCallParams

_ToolSchemaT = TypeVar("_ToolSchemaT")


class BaseCallKwargs(BaseCallParams, Generic[_ToolSchemaT]):
    tools: NotRequired[list[_ToolSchemaT]]



from mirascope.core import openai


@openai.call(model="gpt-4o-mini")
def recommend_book(genre: str):
    """Recommend a {genre} book."""


recommendation = recommend_book("fantasy")
# `recommendation` is properly typed as `OpenAICallResponse`
print(f"Content: {recommendation.content}")
# `recommendation.response` is properly typed as `ChatCompletion`
print(f"Original Response: {recommendation.response}")



from mirascope.core import BasePrompt, prompt_template


@prompt_template(
    "I just read this book: {book:image(detail=high)}. What should I read next?"
)
class BookRecommendationPrompt(BasePrompt):
    book: str | bytes


url = "https://upload.wikimedia.org/wikipedia/en/5/56/TheNameoftheWind_cover.jpg"
prompt = BookRecommendationPrompt(book=url)
print(prompt.message_params())
# > [BaseMessageParam(
#       role='user',
#       content=[
#           ...
#           ImagePart(
#               ...
#               detail='high'
#           ),
#           ...
#       ]
#   )]



"""This module contains the type definition for the Azure call keyword arguments."""

from azure.ai.inference.models import ChatCompletionsToolDefinition, ChatRequestMessage

from ..base import BaseCallKwargs
from .call_params import AzureCallParams


class AzureCallKwargs(AzureCallParams, BaseCallKwargs[ChatCompletionsToolDefinition]):
    model: str
    messages: list[ChatRequestMessage]



"""
Adapted from the Pydantic team's original script.
https://github.com/pydantic/pydantic/blob/main/docs/plugins/griffe_doclinks.py
"""

from __future__ import annotations

import ast
import logging
import re
import sys
import traceback
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

from griffe import Extension, ObjectNode
from griffe import Object as GriffeObject
from pymdownx.slugs import slugify

DOCS_PATH = Path(__file__).parent.parent
slugifier = slugify(case="lower")
logger = logging.getLogger("griffe_doclinks")
logger.setLevel(logging.DEBUG)


def safe_regex_search(pattern: str, string: str, flags: int = 0) -> re.Match | None:
    try:
        return re.search(pattern, string, flags)
    except re.error as e:
        logger.error(f"Regex error with pattern '{pattern}': {str(e)}")
        return None


def find_heading(content: str, slug: str, file_path: Path) -> tuple[str, int]:
    results = safe_regex_search(r"^#+ (.+)", content, flags=re.M)
    for m in results or []:
        heading = m.group(1)
        h_slug = slugifier(heading, "-")
        if h_slug == slug:
            return heading, m.end()
    raise ValueError(f"heading with slug {slug!r} not found in {file_path}")


def get_local_api_link(obj_path: str) -> str:
    path_parts = obj_path.split(".")
    return f"/api/{'/'.join(path_parts)}"  # Remove '.md' and make it relative to root


def insert_or_update_api_section(file_path: Path, api_link: str, obj_path: str) -> None:
    try:
        content = file_path.read_text()
        api_section = (
            f'??? api "API Documentation"\n\n    [`{obj_path}`]({api_link})\n\n'
        )

        if '??? api "API Documentation"' in content:
            content = re.sub(
                r"??? api \"API Documentation\"\n\n.*?\n\n",
                api_section,
                content,
                flags=re.DOTALL,
            )
        else:
            first_heading_match = safe_regex_search(r"^# .+\n", content, re.MULTILINE)
            if first_heading_match:
                first_heading_end = first_heading_match.end()
                content = f"{content[:first_heading_end]}\n{api_section}{content[first_heading_end:]}"
            else:
                content = f"{api_section}{content}"

        file_path.write_text(content)
    except Exception:
        # logger.error(f"Error in insert_or_update_api_section: {str(e)}")
        logger.debug(traceback.format_exc())


def update_links(obj: GriffeObject) -> None:
    try:
        docstring = obj.docstring
        if not docstring or not docstring.value:
            return

        logger.debug(f"Processing docstring for {obj.path}")
        logger.debug(f"Docstring content: {docstring.value}")

        usage_docs_match = safe_regex_search(
            r"usage[\s-]*docs:[\s]*(\S+)", docstring.value, flags=re.I
        )
        if not usage_docs_match:
            logger.debug(f"No usage docs link found in docstring for {obj.path}")
            return

        usage_docs_link = usage_docs_match.group(1)
        logger.debug(f"Found usage docs link: {usage_docs_link}")

        # Parse the usage docs link
        parsed_link = urlparse(usage_docs_link)

        if parsed_link.netloc == "docs.mirascope.io":
            local_link = parsed_link.path.lstrip("/")
        else:
            local_link = usage_docs_link

        # Split the local link into path and fragment
        local_path, _, fragment = local_link.partition("#")

        usage_file_path = DOCS_PATH / local_path.rstrip("/")
        if not usage_file_path.exists():
            usage_file_path = usage_file_path.with_suffix(".md")

        if not usage_file_path.exists():
            logger.warning(f"Usage docs file not found: {usage_file_path}")
            return

        local_api_link = get_local_api_link(obj.path)
        insert_or_update_api_section(usage_file_path, local_api_link, obj.path)

        # Reconstruct the link with the fragment
        full_local_link = f"/{local_path}"
        if fragment:
            full_local_link += f"#{fragment}"

        # Create the usage docs section
        usage_docs_section = f"??? abstract \"Usage Documentation\"\n    [{usage_file_path.stem.replace('_', ' ').title()}]({full_local_link})\n\n"

        # Replace the original usage docs line with the new section
        docstring.value = re.sub(
            r"usage[\s-]*docs:[\s]*\S+\n?",
            usage_docs_section,
            docstring.value,
            flags=re.I,
        )

        logger.debug(f"Successfully updated links for {obj.path}")
    except Exception as e:
        logger.error(f"Error in update_links for {obj.path}: {str(e)}")
        logger.debug(traceback.format_exc())


class UpdateDocstringsExtension(Extension):
    def on_instance(  # type: ignore
        self, *, node: ast.AST | ObjectNode, obj: GriffeObject, agent: Any
    ) -> None:
        try:
            if not obj.is_alias and obj.docstring is not None:
                update_links(obj)
        except Exception as e:
            logger.error(f"Error in UpdateDocstringsExtension for {obj.path}: {str(e)}")
            logger.debug(traceback.format_exc())


# Add a hook to catch and log any unhandled exceptions
def exception_handler(exctype: Any, value: Any, tb: Any) -> None:
    logger.error("Unhandled exception:", exc_info=(exctype, value, tb))


sys.excepthook = exception_handler



"""This module defines the base class for tools used in LLM calls.

usage-docs: learn/tools.md#the-basetool-class
"""

from __future__ import annotations

import inspect
import warnings
from abc import abstractmethod
from collections.abc import Callable
from typing import Any, ClassVar, Generic, TypeVar

from pydantic import BaseModel, ConfigDict
from pydantic.json_schema import (
    DEFAULT_REF_TEMPLATE,
    GenerateJsonSchema,
    JsonSchemaMode,
    JsonSchemaValue,
)
from pydantic_core.core_schema import CoreSchema
from typing_extensions import TypedDict

from . import _utils

_BaseToolT = TypeVar("_BaseToolT", bound=BaseModel)
_ToolSchemaT = TypeVar("_ToolSchemaT")


class ToolConfig(TypedDict, total=False):
    """A base class for tool configurations."""


class GenerateJsonSchemaNoTitles(GenerateJsonSchema):
    _openai_strict: ClassVar[bool] = False

    def _remove_title(self, obj: Any) -> Any:  # noqa: ANN401
        if isinstance(obj, dict):
            if self._openai_strict and "type" in obj and obj["type"] == "object":
                obj["additionalProperties"] = False
            if "type" in obj or "$ref" in obj or "properties" in obj:
                obj.pop("title", None)

            for key, value in list(obj.items()):
                obj[key] = self._remove_title(value)
        elif isinstance(obj, list):
            return [self._remove_title(item) for item in obj]

        return obj

    def generate(
        self, schema: CoreSchema, mode: JsonSchemaMode = "validation"
    ) -> JsonSchemaValue:
        json_schema = super().generate(schema, mode=mode)
        json_schema.pop("title", None)
        json_schema.pop("description", None)
        json_schema = self._remove_title(json_schema)
        return json_schema


class BaseTool(BaseModel, Generic[_ToolSchemaT]):
    '''A class for defining tools for LLM calls.

    Example:

    ```python
    from mirascope.core import BaseTool
    from pydantic import Field


    class FormatBook(BaseTool):
        """Returns a nicely formatted book recommendation."""

        title: str = Field(..., description="The title of the book.")
        author: str = Field(..., description="The author of the book.")

        def call(self) -> str:
            return f"{self.title} by {self.author}"
    ```
    '''

    __provider__: ClassVar[str] = "NONE"
    __tool_config_type__: ClassVar[type[ToolConfig]] = ToolConfig
    __custom_name__: ClassVar[str] = ""
    tool_config: ClassVar[ToolConfig] = ToolConfig()
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def _name(cls) -> str:
        """Returns the name of the tool."""
        return cls.__custom_name__ if cls.__custom_name__ else cls.__name__

    @classmethod
    def _description(cls) -> str:
        """Returns the description of the tool."""
        return (
            inspect.cleandoc(cls.__doc__)
            if cls.__doc__
            else _utils.DEFAULT_TOOL_DOCSTRING
        )

    @property
    def args(self) -> dict[str, Any]:
        """The arguments of the tool as a dictionary."""
        return {
            field: getattr(self, field)
            for field in self.model_fields
            if field != "tool_call"
        }

    @abstractmethod
    def call(self) -> Any:  # noqa: ANN401
        """The method to call the tool."""
        ...

    @classmethod
    def type_from_fn(cls: type[_BaseToolT], fn: Callable) -> type[_BaseToolT]:
        """Returns this tool type converted from a function.

        Args:
            fn: The function to convert into this tool type.
        """
        return _utils.convert_function_to_base_tool(fn, cls)

    @classmethod
    def type_from_base_model_type(
        cls: type[_BaseToolT], tool_type: type[BaseModel]
    ) -> type[_BaseToolT]:
        """Returns this tool type converted from a given base tool type.

        Args:
            tool_type: The tool type to convert into this tool type. This can be a
                custom `BaseTool` or `BaseModel` definition.
        """
        return _utils.convert_base_model_to_base_tool(tool_type, cls)

    @classmethod
    def type_from_base_type(
        cls: type[_BaseToolT], base_type: type[_utils.BaseType]
    ) -> type[_BaseToolT]:
        """Returns this tool type converted from a base type.

        Args:
            base_type: The base type (e.g. `int`) to convert into this tool type.
        """
        return _utils.convert_base_type_to_base_tool(base_type, cls)

    @classmethod
    def tool_schema(cls) -> _ToolSchemaT:
        raise RuntimeError(
            f"{cls.__name__}.tool_schema() is not implemented. "
            "This method should be implemented in provider-specific tool classes."
        )

    @classmethod
    def model_json_schema(
        cls,
        by_alias: bool = True,
        ref_template: str = DEFAULT_REF_TEMPLATE,
        schema_generator: type[GenerateJsonSchema] = GenerateJsonSchemaNoTitles,
        mode: JsonSchemaMode = "validation",
    ) -> dict[str, Any]:
        """Returns the generated JSON schema for the class."""
        cls.warn_for_unsupported_configurations()
        return super().model_json_schema(
            by_alias=by_alias,
            ref_template=ref_template,
            schema_generator=schema_generator,
            mode=mode,
        )

    @classmethod
    def warn_for_unsupported_configurations(cls) -> None:
        """Warns when a specific provider does not support provided config options."""
        unsupported_tool_keys = _utils.get_unsupported_tool_config_keys(
            cls.tool_config, cls.__tool_config_type__
        )
        if unsupported_tool_keys:
            warnings.warn(
                f"{cls.__provider__} does not support the following tool "
                f"configurations, so they will be ignored: {unsupported_tool_keys}",
                UserWarning,
            )

        if "strict" in cls.model_config and cls.__provider__ not in ["openai", "azure"]:
            warnings.warn(
                f"{cls.__provider__} does not support strict structured outputs, but "
                "you have configured `strict=True` in your `ResponseModelConfigDict`. "
                "Ignoring `strict` as this feature is only supported by OpenAI.",
                UserWarning,
            )



from collections.abc import AsyncGenerator, Generator, Iterable, Iterator
from typing import Any

import pytest

from mirascope.core.base._utils._get_create_fn_or_async_create_fn import (
    get_async_create_fn,
    get_create_fn,
)


@pytest.mark.asyncio
async def test_get_async_create_fn_non_streaming():
    async def async_func(**kwargs: Any) -> str:
        return "non-streaming result"

    create_or_stream = get_async_create_fn(async_func)
    result = await create_or_stream(stream=False)
    assert result == "non-streaming result"


@pytest.mark.asyncio
async def test_get_async_create_fn_streaming_with_generator():
    async def async_func(**kwargs: Any) -> str:
        return "non-streaming result"  # pragma: no cover

    async def async_generator_func(**kwargs: Any) -> AsyncGenerator[str, None]:
        yield "streaming result 1"
        yield "streaming result 2"

    create_or_stream = get_async_create_fn(async_func, async_generator_func)
    stream_result = await create_or_stream(stream=True)
    assert isinstance(stream_result, AsyncGenerator)

    results = [item async for item in stream_result]
    assert results == ["streaming result 1", "streaming result 2"]


@pytest.mark.asyncio
async def test_get_async_create_fn_streaming_without_generator():
    async def async_func(**kwargs: Any) -> AsyncGenerator[str, None]:
        async def generator() -> AsyncGenerator[str, None]:
            yield "streaming result 1"
            yield "streaming result 2"

        return generator()

    create_or_stream = get_async_create_fn(async_func)
    stream_result = await create_or_stream(stream=True)
    assert isinstance(stream_result, AsyncGenerator)

    results = [item async for item in stream_result]
    assert results == ["streaming result 1", "streaming result 2"]


@pytest.mark.asyncio
async def test_get_async_create_fn_with_additional_kwargs():
    async def async_func(**kwargs: Any) -> str:
        return f"result: {kwargs.get('param', '')}"

    create_or_stream = get_async_create_fn(async_func)
    result = await create_or_stream(stream=False, param="test")
    assert result == "result: test"


@pytest.mark.asyncio
async def test_get_async_create_fn_streaming_with_additional_kwargs():
    async def async_func(**kwargs: Any) -> str:
        return "unused"  # pragma: no cover

    async def async_generator_func(**kwargs: Any) -> AsyncGenerator[str, None]:
        yield f"streaming result: {kwargs.get('param', '')}"

    create_or_stream = get_async_create_fn(async_func, async_generator_func)
    stream_result = await create_or_stream(stream=True, param="test")
    assert isinstance(stream_result, AsyncGenerator)

    results = [item async for item in stream_result]
    assert results == ["streaming result: test"]


@pytest.mark.asyncio
async def test_get_async_create_fn_default_stream_value():
    async def async_func(**kwargs: Any) -> str:
        return "non-streaming result"

    create_or_stream = get_async_create_fn(async_func)
    result = await create_or_stream()  # Default stream=False
    assert result == "non-streaming result"


@pytest.mark.asyncio
async def test_get_async_create_fn_type_casting():
    async def async_func(**kwargs: Any) -> str:
        return "non-streaming result"

    async def async_generator_func(**kwargs: Any) -> AsyncGenerator[str, None]:
        yield "streaming result"

    create_or_stream = get_async_create_fn(async_func, async_generator_func)

    # Test non-streaming mode
    non_stream_result = await create_or_stream(stream=False)
    assert isinstance(non_stream_result, str)
    assert non_stream_result == "non-streaming result"

    # Test streaming mode
    stream_result = await create_or_stream(stream=True)
    assert isinstance(stream_result, AsyncGenerator)

    # Consume the AsyncGenerator to verify its content
    results = [item async for item in stream_result]
    assert results == ["streaming result"]


def test_get_create_fn_non_streaming():
    def sync_func(**kwargs: Any) -> str:
        return "non-streaming result"

    create_or_stream = get_create_fn(sync_func)
    result = create_or_stream(stream=False)
    assert result == "non-streaming result"


def test_get_create_fn_streaming_with_generator():
    def sync_func(**kwargs: Any) -> str:
        return "non-streaming result"  # pragma: no cover

    def sync_generator_func(**kwargs: Any) -> Iterator[str]:
        yield "streaming result 1"
        yield "streaming result 2"

    create_or_stream = get_create_fn(sync_func, sync_generator_func)
    stream_result = create_or_stream(stream=True)
    assert isinstance(stream_result, Generator)

    results = list(stream_result)
    assert results == ["streaming result 1", "streaming result 2"]


def test_get_create_fn_streaming_without_generator():
    def sync_func(**kwargs: Any) -> Iterator[str]:
        yield "streaming result 1"
        yield "streaming result 2"

    create_or_stream = get_create_fn(sync_func)
    stream_result = create_or_stream(stream=True)
    assert isinstance(stream_result, Generator)

    results = list(stream_result)
    assert results == ["streaming result 1", "streaming result 2"]


def test_get_create_fn_streaming_with_iterable():
    def sync_func(**kwargs: Any) -> Iterable[str]:
        return ["streaming result 1", "streaming result 2"]

    create_or_stream = get_create_fn(sync_func)
    stream_result = create_or_stream(stream=True)
    assert isinstance(stream_result, Generator)

    results = list(stream_result)
    assert results == ["streaming result 1", "streaming result 2"]


def test_get_create_fn_with_additional_kwargs():
    def sync_func(**kwargs: Any) -> str:
        return f"result: {kwargs.get('param', '')}"

    create_or_stream = get_create_fn(sync_func)
    result = create_or_stream(stream=False, param="test")
    assert result == "result: test"


def test_get_create_fn_streaming_with_additional_kwargs():
    def sync_generator_func(**kwargs: Any) -> Iterator[str]:
        yield f"streaming result: {kwargs.get('param', '')}"

    create_or_stream = get_create_fn(lambda **kwargs: "unused", sync_generator_func)
    stream_result = create_or_stream(stream=True, param="test")
    assert isinstance(stream_result, Generator)

    results = list(stream_result)
    assert results == ["streaming result: test"]


def test_get_create_fn_default_stream_value():
    def sync_func(**kwargs: Any) -> str:
        return "non-streaming result"

    create_or_stream = get_create_fn(sync_func)
    result = create_or_stream()  # Default stream=False
    assert result == "non-streaming result"


def test_get_create_fn_type_casting():
    def sync_func(**kwargs: Any) -> str:
        return "non-streaming result"

    def sync_generator_func(**kwargs: Any) -> Iterator[str]:
        yield "streaming result"

    create_or_stream = get_create_fn(sync_func, sync_generator_func)

    # Test non-streaming mode
    non_stream_result = create_or_stream(stream=False)
    assert isinstance(non_stream_result, str)
    assert non_stream_result == "non-streaming result"

    # Test streaming mode
    stream_result = create_or_stream(stream=True)
    assert isinstance(stream_result, Generator)

    # Consume the Generator to verify its content
    results = list(stream_result)
    assert results == ["streaming result"]



"""This module defines the function return type for functions as LLM calls.

usage docs: learn/dynamic_configuration.md#dynamic-configuration-options
"""

from groq.types.chat import ChatCompletionMessageParam

from ..base import BaseDynamicConfig
from .call_params import GroqCallParams

GroqDynamicConfig = BaseDynamicConfig[ChatCompletionMessageParam, GroqCallParams]
"""The function return type for functions wrapped with the `groq_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.groq import GroqDynamicConfig, groq_call


@groq_call("llama-3.1-8b-instant")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> GroqDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



"""OpenAI modules for the v0 look-alike implementation."""

from typing import ClassVar, Generic, TypeVar

from openai.types.chat import ChatCompletionToolChoiceOptionParam
from openai.types.chat.completion_create_params import ResponseFormat
from pydantic import ConfigDict

from ..core.openai import (
    OpenAICallResponse,
    OpenAICallResponseChunk,
    OpenAITool,
    openai_call,
)
from .base import BaseCall, BaseCallParams, BaseExtractor, ExtractedType


class OpenAICallParams(BaseCallParams):
    """The parameters to use when calling the OpenAI API."""

    model: str
    frequency_penalty: float | None = None
    logit_bias: dict[str, int] | None = None
    logprobs: bool | None = None
    max_tokens: int | None = None
    n: int | None = None
    presence_penalty: float | None = None
    response_format: ResponseFormat | None = None
    seed: int | None = None
    stop: str | list[str] | None = None
    temperature: float | None = None
    tool_choice: ChatCompletionToolChoiceOptionParam | None = None
    top_logprobs: int | None = None
    top_p: float | None = None
    user: str | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")


class OpenAICall(BaseCall[OpenAICallResponse, OpenAICallResponseChunk, OpenAITool]):
    call_params: ClassVar[BaseCallParams] = OpenAICallParams(model="gpt-4o-mini")

    _decorator = openai_call
    _provider = "openai"


T = TypeVar("T", bound=ExtractedType)


class OpenAIExtractor(BaseExtractor[T], Generic[T]):
    call_params: ClassVar[BaseCallParams] = OpenAICallParams(model="gpt-4o-mini")

    _decorator = openai_call
    _provider = "openai"



"""Tests the `openai.call` module."""

import sys
from unittest.mock import MagicMock, patch

from mirascope.core.openai import _utils
from mirascope.core.openai.call_params import OpenAICallParams
from mirascope.core.openai.call_response import OpenAICallResponse
from mirascope.core.openai.call_response_chunk import OpenAICallResponseChunk
from mirascope.core.openai.dynamic_config import OpenAIDynamicConfig
from mirascope.core.openai.stream import OpenAIStream
from mirascope.core.openai.tool import OpenAITool


def test_openai_call() -> None:
    """Tests the `openai_call` decorator."""

    if "mirascope.core.openai._call" in sys.modules:
        del sys.modules["mirascope.core.openai._call"]

    with patch(
        "mirascope.core.base.call_factory", new_callable=MagicMock
    ) as mock_call_factory:
        import mirascope.core.openai._call  # noqa: F401

        mock_call_factory.assert_called_once_with(
            TCallResponse=OpenAICallResponse,
            TCallResponseChunk=OpenAICallResponseChunk,
            TDynamicConfig=OpenAIDynamicConfig,
            TToolType=OpenAITool,
            TStream=OpenAIStream,
            TCallParams=OpenAICallParams,
            default_call_params={},
            setup_call=_utils.setup_call,
            get_json_output=_utils.get_json_output,
            handle_stream=_utils.handle_stream,
            handle_stream_async=_utils.handle_stream_async,
        )



from pydantic import BaseModel, Field

from mirascope.core import BasePrompt, anthropic, openai, prompt_template


class Eval(BaseModel):
    score: float = Field(..., description="A score between [0.0, 5.0]")
    reasoning: str = Field(
        ..., description="The reasoning for the score in 100 characters or less."
    )


@prompt_template(
    """
    Evaluate the toxicity of the following text on a scale from 0 to 5:

    0 - No toxicity
    1 - Mild toxicity; minor offensive elements or insensitivity
    2 - Moderate toxicity; some harmful or disrespectful language
    3 - High toxicity; explicit and aggressive harmful content
    4 - Severe toxicity; dangerous, hateful, or extremist content
    5 - Extreme toxicity; could cause significant harm

    Text to evaluate: {text}
    """
)
class ToxicityEvaluationPrompt(BasePrompt):
    text: str


toxic_input = "Why even bother trying? With your laziness and abilities, it's probably not even possible anyway."
prompt = ToxicityEvaluationPrompt(text=toxic_input)

judges = [
    openai.call("gpt-4o", response_model=Eval),
    anthropic.call("claude-3-5-sonnet-20240620", response_model=Eval),
]

evaluations: list[Eval] = [prompt.run(judge) for judge in judges]

for evaluation in evaluations:
    print(evaluation)
# Output:
# score=3.0 reasoning='The text is explicitly demeaning and discourages effort, which is harmful.'
# score=2.5 reasoning='Discouraging, insulting laziness, and implying inability. Promotes negativity and is demeaning.'



"""Tests the `gemini._utils.convert_message_params` function."""

import io
from unittest.mock import MagicMock, patch

import pytest
from google.generativeai.types import ContentDict

from mirascope.core.base import (
    AudioPart,
    BaseMessageParam,
    CacheControlPart,
    ImagePart,
    TextPart,
)
from mirascope.core.gemini._utils._convert_message_params import convert_message_params


@patch("PIL.Image.open", new_callable=MagicMock)
def test_convert_message_params(mock_image_open: MagicMock) -> None:
    """Tests the `convert_message_params` function."""
    mock_image_open.return_value = "test"
    message_params: list[BaseMessageParam | ContentDict] = [
        BaseMessageParam(role="system", content="You are a helpful assistant."),
        BaseMessageParam(role="user", content="Hello"),
        {"role": "user", "parts": ["Hello", {"type": "text", "text": "Hello"}]},
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="test"),
                ImagePart(
                    type="image", media_type="image/jpeg", image=b"image", detail=None
                ),
                AudioPart(type="audio", media_type="audio/wav", audio=b"audio"),
            ],
        ),
    ]
    converted_message_params = convert_message_params(message_params)
    assert converted_message_params == [
        {"role": "user", "parts": ["You are a helpful assistant."]},
        {"role": "model", "parts": ["Ok! I will adhere to this system message."]},
        {"role": "user", "parts": ["Hello"]},
        {"role": "user", "parts": ["Hello", {"type": "text", "text": "Hello"}]},
        {
            "role": "user",
            "parts": ["test", "test", {"mime_type": "audio/wav", "data": b"audio"}],
        },
    ]
    mock_image_open.assert_called_once()
    bytes_io: io.BytesIO = mock_image_open.call_args.args[0]
    assert bytes_io.getvalue() == b"image"

    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg. Gemini currently only supports "
        "JPEG, PNG, WebP, HEIC, and HEIF images.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail=None,
                        )
                    ],
                ),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: audio/unknown. "
        "Gemini currently only supports WAV, MP3, AIFF, AAC, OGG, "
        "and FLAC audio file types.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(
                            type="audio", media_type="audio/unknown", audio=b"audio"
                        )
                    ],
                ),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Gemini currently only supports text, image, and audio parts. Part provided: cache_control",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        CacheControlPart(type="cache_control", cache_type="ephemeral")
                    ],
                )
            ]
        )



import asyncio
from collections.abc import Callable
from enum import StrEnum

from mirascope.core import BasePrompt, openai, prompt_template


class Audience(StrEnum):
    general = "general"
    professional = "professional"
    academic = "academic"
    friendly = "friendly"
    formal = "formal"


class Tone(StrEnum):
    neutral = "neutral"
    positive = "positive"
    negative = "negative"
    professional = "professional"
    casual = "casual"


@prompt_template(
    """
    SYSTEM:
    You are a professional translator who is a native Japanese speaker.
    Translate the English text into natural Japanese without changing the content.
    Please make sure that the text is easy to read for everyone by using appropriate paragraphs so that it does not become difficult to read.

    Please consider the following parameters in your translation:

    tone: {tone}
    The tone can be one of the following:
    - neutral: Maintain a balanced and impartial tone
    - positive: Use upbeat and optimistic language
    - negative: Convey a more critical or pessimistic view
    - professional: Use formal and business-appropriate language
    - casual: Use informal, conversational language

    audience: {audience}
    The audience can be one of the following:
    - general: For the general public, use common terms and clear explanations
    - professional: For industry professionals, use appropriate jargon and technical terms
    - academic: For scholarly contexts, use formal language and cite sources if necessary
    - friendly: For casual, familiar contexts, use warm and approachable language
    - formal: For official or ceremonial contexts, use polite and respectful language

    Adjust your translation style based on the specified tone and audience to ensure the message is conveyed appropriately.

    USER:
    text: {text}
    """
)
class ParametrizedTranslatePrompt(BasePrompt):
    tone: Tone
    audience: Audience
    text: str

    async def translate(self, call: Callable, model: str) -> str:
        response = await self.run_async(call(model))
        return response.content


text = """
The old codger, a real bootstrapper, had been burning the candle at both ends trying to make his pie-in-the-sky business idea fly. He'd been spinning his wheels for months, barking up the wrong tree with his half-baked marketing schemes.
"""
prompt = ParametrizedTranslatePrompt(
    text=text, tone=Tone.negative, audience=Audience.general
)
parametrized_translation = asyncio.run(
    prompt.translate(call=openai.call, model="gpt-4o-mini")
)
print(f"Parametrized with translation: {parametrized_translation}")



from typing import Annotated

from pydantic import AfterValidator, ValidationError

from mirascope.core import BaseTool, openai, prompt_template


def is_upper(v: str) -> str:
    assert v.isupper(), "Must be uppercase"
    return v


class FormatBook(BaseTool):
    title: Annotated[str, AfterValidator(is_upper)]
    author: str

    def call(self) -> str:
        return f"{self.title} by {self.author}"


@openai.call(model="gpt-4", tools=[FormatBook])
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


try:
    response = recommend_book("fantasy")
    if tool := response.tool:
        print(tool.call())
    else:
        print(response.content)
except ValidationError as e:
    print(e)
    # > 1 validation error for FormatBook
    #   title
    #     Assertion failed, Must be uppercase [type=assertion_error, input_value='The Name of the Wind', input_type=str]
    #       For further information visit https://errors.pydantic.dev/2.8/v/assertion_error



"""This module defines the function return type for functions as LLM calls.

usage docs: learn/dynamic_configuration.md#dynamic-configuration-options
"""

from openai.types.chat import ChatCompletionMessageParam

from ..base import BaseDynamicConfig
from .call_params import OpenAICallParams

OpenAIDynamicConfig = BaseDynamicConfig[ChatCompletionMessageParam, OpenAICallParams]
"""The function return type for functions wrapped with the `openai_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.openai import OpenAIDynamicConfig, openai_call


@openai_call("gpt-4o-mini")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> OpenAIDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



from mirascope.core import openai, prompt_template
from mirascope.integrations.otel import configure, with_otel

configure()


@with_otel()
@openai.call(
    model="gpt-4o-mini",
    stream=True,
    call_params={"stream_options": {"include_usage": True}},
)
@prompt_template("Recommend a {genre} book.")
def recommend_book(genre: str): ...


for chunk, _ in recommend_book("fantasy"):
    print(chunk.content, end="", flush=True)



"""Tests the `mistral._utils.setup_call` module."""

from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from mistralai.async_client import MistralAsyncClient
from mistralai.models.chat_completion import (
    ChatCompletionResponse,
    ChatCompletionStreamResponse,
    ChatMessage,
    ToolChoice,
)

from mirascope.core.mistral._utils._setup_call import setup_call
from mirascope.core.mistral.tool import MistralTool


@pytest.fixture()
def mock_base_setup_call() -> MagicMock:
    """Returns the mock setup_call function."""
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = [MagicMock() for _ in range(3)] + [{}]
    return mock_setup_call


@patch(
    "mirascope.core.mistral._utils._setup_call.MistralClient.chat_stream",
    return_value=MagicMock(),
)
@patch(
    "mirascope.core.mistral._utils._setup_call.MistralClient.chat",
    return_value=MagicMock(),
)
@patch(
    "mirascope.core.mistral._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.mistral._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_mistral_chat: MagicMock,
    mock_mistral_chat_stream: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    mock_utils.setup_call = mock_base_setup_call
    mock_chat_iterator = MagicMock()
    mock_chat_iterator.__iter__.return_value = ["chat"]
    mock_mistral_chat_stream.return_value = mock_chat_iterator
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="mistral-large-latest",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "model" in call_kwargs and call_kwargs["model"] == "mistral-large-latest"
    assert "messages" in call_kwargs and call_kwargs["messages"] == messages
    mock_base_setup_call.assert_called_once_with(fn, {}, None, None, MistralTool, {})
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )
    assert messages == mock_convert_message_params.return_value
    create(stream=False, **call_kwargs)
    mock_mistral_chat.assert_called_once_with(**call_kwargs)
    stream = create(stream=True, **call_kwargs)
    mock_mistral_chat_stream.assert_called_once_with(**call_kwargs)
    assert next(stream) == "chat"  # pyright: ignore [reportArgumentType]


@patch(
    "mirascope.core.mistral._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.mistral._utils._setup_call._utils", new_callable=MagicMock)
@pytest.mark.asyncio
async def test_async_setup_call(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    mock_mistral_chat = AsyncMock(spec=ChatCompletionResponse)
    mock_mistral_chat.__name__ = "chat"

    mock_stream_response = AsyncMock(spec=ChatCompletionStreamResponse)
    mock_stream_response.text = "chat"

    class AsyncMockIterator:
        def __init__(self, item):
            self.item = iter(item)

        def __aiter__(self):
            return self

        async def __anext__(self):
            try:
                return next(self.item)
            except StopIteration:
                raise StopAsyncIteration

    mock_iterator = AsyncMockIterator([mock_stream_response])

    mock_client = AsyncMock(spec=MistralAsyncClient, name="mock_client")
    mock_client.chat_stream.return_value = mock_iterator
    mock_client.chat.return_value = mock_mistral_chat

    mock_utils.setup_call = mock_base_setup_call

    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="mistral-large-latest",
        client=mock_client,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "model" in call_kwargs and call_kwargs["model"] == "mistral-large-latest"
    assert "messages" in call_kwargs and call_kwargs["messages"] == messages
    mock_base_setup_call.assert_called_once_with(fn, {}, None, None, MistralTool, {})
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )

    mock_mistral_chat.return_value = MagicMock(spec=ChatCompletionResponse)
    chat = await create(stream=False, **call_kwargs)
    stream = await create(stream=True, **call_kwargs)
    result = []
    async for chunk in stream:
        result.append(chunk)
    assert result == [mock_stream_response]
    assert isinstance(chat, ChatCompletionResponse)
    assert isinstance(stream, AsyncMockIterator)


@patch(
    "mirascope.core.mistral._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.mistral._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with JSON mode."""
    mock_utils.setup_call = mock_base_setup_call
    mock_json_mode_content = MagicMock()
    mock_json_mode_content.return_value = "\n\njson_mode_content"
    mock_utils.json_mode_content = mock_json_mode_content
    mock_base_setup_call.return_value[1] = [ChatMessage(role="user", content="test")]
    mock_base_setup_call.return_value[-1]["tools"] = MagicMock()
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="mistral-large-latest",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        extract=False,
    )
    assert messages[-1].content == "test\n\njson_mode_content"
    assert "tools" not in call_kwargs

    mock_base_setup_call.return_value[1] = [
        ChatMessage(role="assistant", content="test"),
    ]
    _, _, messages, _, call_kwargs = setup_call(
        model="mistral-large-latest",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        extract=False,
    )
    assert messages[-1] == ChatMessage(role="user", content="json_mode_content")


@patch(
    "mirascope.core.mistral._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.mistral._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_extract(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with extraction."""
    mock_utils.setup_call = mock_base_setup_call
    _, _, _, _, call_kwargs = setup_call(
        model="mistral-large-latest",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=True,
    )
    assert "tool_choice" in call_kwargs and call_kwargs["tool_choice"] == ToolChoice.any



"""Tests the `_utils.setup_extract_tool` function."""

from pydantic import BaseModel

from mirascope.core.base._utils._setup_extract_tool import setup_extract_tool
from mirascope.core.base.tool import BaseTool


def test_setup_extract_tool() -> None:
    """Tests the `_utils.setup_extract_tool` function."""

    class Book(BaseModel):
        title: str

    tool_type = setup_extract_tool(Book, BaseTool)
    assert tool_type.__name__ == "Book"
    assert tool_type.__base__ == Book
    assert tool_type.__bases__ == (Book, BaseTool)

    tool_type = setup_extract_tool(str, BaseTool)
    assert tool_type.__name__ == "str"
    assert tool_type.__base__ == BaseTool
    assert tool_type.__bases__ == (BaseTool,)



from mirascope.core import BasePrompt, prompt_template


@prompt_template(
    """
    Recommend a book from one of the following genres:
    {genres:list}

    Examples:
    {examples:lists}
    """
)
class BookRecommendationPrompt(BasePrompt):
    genres: list[str]
    examples: list[tuple[str, str]]


prompt = BookRecommendationPrompt(
    genres=["fantasy", "scifi", "mystery"],
    examples=[
        ("Title: The Name of the Wind", "Author: Patrick Rothfuss"),
        ("Title: Mistborn: The Final Empire", "Author: Brandon Sanderson"),
    ],
)
print(prompt)
# > Recommend a book from one of the following genres:
#   fantasy
#   scifi
#   mystery
#
#   Examples:
#   Title: The Name of the Wind
#   Author: Patrick Rothfuss
#
#   Title: Mistborn: The Final Empire
#   Author: Brandon Sanderson



"""The `GroqTool` class for easy tool usage with Groq LLM calls.

usage docs: learn/tools.md#using-tools-with-standard-calls
"""

from __future__ import annotations

import jiter
from groq.types.chat import (
    ChatCompletionMessageToolCall,
    ChatCompletionToolParam,
)
from groq.types.shared_params import FunctionDefinition
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool


class GroqTool(BaseTool[ChatCompletionToolParam]):
    """A class for defining tools for Groq LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.groq import groq_call


    def format_book(title: str, author: str) -> str:
        return f"{title} by {author}"


    @groq_call("llama-3.1-8b-instant", tools=[format_book])
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    response = recommend_book("fantasy")
    if tool := response.tool:  # returns an `GroqTool` instance
        print(tool.call())
    ```
    """

    __provider__ = "groq"

    tool_call: SkipJsonSchema[ChatCompletionMessageToolCall]

    @classmethod
    def tool_schema(cls) -> ChatCompletionToolParam:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined.

        Example:
        ```python
        from mirascope.core.groq import GroqTool


        def format_book(title: str, author: str) -> str:
            return f"{title} by {author}"


        tool_type = GroqTool.type_from_fn(format_book)
        print(tool_type.tool_schema())  # prints the Groq-specific tool schema
        ```
        """
        fn = FunctionDefinition(name=cls._name(), description=cls._description())
        model_schema = cls.model_json_schema()
        if model_schema["properties"]:
            fn["parameters"] = model_schema
        return ChatCompletionToolParam(function=fn, type="function")

    @classmethod
    def from_tool_call(cls, tool_call: ChatCompletionMessageToolCall) -> GroqTool:
        """Constructs an `GroqTool` instance from a `tool_call`.

        Args:
            tool_call: The Groq tool call from which to construct this tool instance.
        """
        model_json = jiter.from_json(tool_call.function.arguments.encode())
        model_json["tool_call"] = tool_call.model_dump()
        return cls.model_validate(model_json)



def format_book(title: str, author: str) -> str:
    """Format a book recommendation.

    Example:
        {"title": "THE HOBBIT", "author": "J.R.R. Tolkien"}

    Example:
        {"title": "THE NAME OF THE WIND", "author": "Patrick Rothfuss"}
    """
    return f"I recommend {title} by {author}."



import logfire
from pydantic import BaseModel

from mirascope.core import openai, prompt_template
from mirascope.integrations.logfire import with_logfire

logfire.configure(pydantic_plugin=logfire.PydanticPlugin(record="all"))


class Book(BaseModel):
    title: str
    author: str


@with_logfire()
@openai.call(model="gpt-4o-mini", response_model=Book)
@prompt_template("Recommend a {genre} book.")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



"""Tests the `_utils.convert_base_model_to_base_tool` module."""

from pydantic import BaseModel

from mirascope.core.base._utils._convert_base_model_to_base_tool import (
    convert_base_model_to_base_tool,
)
from mirascope.core.base._utils._default_tool_docstring import DEFAULT_TOOL_DOCSTRING
from mirascope.core.base.tool import BaseTool


def test_convert_base_model_to_base_tool() -> None:
    """Tests the `convert_base_model_to_base_tool` function."""

    class Book(BaseModel):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool = convert_base_model_to_base_tool(Book, BaseTool)
    assert tool._name() == "Book"
    assert tool._description() == DEFAULT_TOOL_DOCSTRING
    assert (
        tool(title="The Name of the Wind", author="Patrick Rothfuss").call()  # type: ignore
        == "The Name of the Wind by Patrick Rothfuss"
    )



import asyncio

from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


class PromptVariations(BaseModel):
    variations: list[str] = Field(..., description="Variations of the original prompt")


@openai.call(model="gpt-4o-mini", response_model=PromptVariations)
@prompt_template(
    """
    Return the {num_prompts} alternate variations of the prompt which retain the
    full meaning but uses different phrasing.
    Prompt: {prompt}
    """
)
def get_prompt_variations(prompt: str, num_prompts: int): ...


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Answer the following question going step by step:
    {query}
    """
)
async def zero_shot_cot(query: str): ...


class ResponseDetails(BaseModel):
    solution_number: int = Field(
        ..., description="The actual number given as the answer in a solution."
    )
    correctness_probability: float = Field(
        ...,
        ge=0,
        le=1,
        description="An estimated probability that the given solution is correct from 0.0 to 1.0",
    )


@openai.call(model="gpt-4o-mini", response_model=ResponseDetails)
@prompt_template(
    """
    Here is a query and a response which attempts to answer the query.
    Prompt: {query}
    Response: {response}

    Extract the raw numerical value of the answer given by the response, and also
    give an estimate between 0.0 and 1.0 of the probability that this solution
    is correct.
    """
)
async def evaluate_response(query: str, response: str): ...


async def diverse(query: str, num_variations: int):
    # Gather the variations of the prompt
    alternate_variations = get_prompt_variations(query, num_variations - 1)
    # Uncomment to see intermediate steps
    # print(alternate_variations)
    all_variations = alternate_variations.variations + [query]

    # Generate a unique reasoning chain for each prompt variation with CoT
    cot_tasks = [zero_shot_cot(prompt) for prompt in all_variations]
    cot_responses = [response.content for response in await asyncio.gather(*cot_tasks)]
    # print(cot_responses)

    # Evaluate each reasoning chain
    eval_tasks = [
        evaluate_response(query, cot_response) for cot_response in cot_responses
    ]
    eval_responses = await asyncio.gather(*eval_tasks)
    # print(eval_responses)

    response_scores = {}
    for eval_response in eval_responses:
        if eval_response.solution_number not in response_scores:
            response_scores[eval_response.solution_number] = 0
        response_scores[eval_response.solution_number] += (
            eval_response.correctness_probability
        )
    # print(response_scores)
    best_response = max(response_scores, key=response_scores.get)  # type: ignore
    return best_response


async def run_self_consistency(prompt, num_variations=3):
    return await diverse(prompt, num_variations)


query = """
A committee of 3 people must be formed from a pool of 6 people, but Amy and Bob do not
get along and should not be on the committee at the same time. How many viable
combinations are there?
"""

print(asyncio.run(run_self_consistency(query)))
# > 16


# Intermediate Steps

# alternate_variations = get_prompt_variations(query, num_variations - 1)
# > variations=['A group of 3 individuals needs to be selected from a total of 6, but Amy and Bob cannot serve on the committee together due to their conflict. What are the possible combinations?', 'We need to establish a committee of 3 members from a selection of 6 individuals, with the stipulation that Amy and Bob must not both be included in the committee. How many suitable combinations can be created?']

# cot_responses = [response.content for response in await asyncio.gather(*cot_tasks)]
# > ["To find the number of possible combinations for selecting a committee of 3 individuals from 6, with the constraint that Amy (A) and Bob (B) cannot be selected together, we can break down the process step by step.\n\n### Step 1: Identify Total Individuals\nWe have the following individuals:\n- A (Amy)\n- B (Bob)\n- C\n- D\n- E\n- F\n\n### Step 2: Calculate Total Combinations Without Constraints\nFirst, we’ll calculate the total number of combinations of 3 individuals from 6 without any restrictions. This can be done using the combination formula \\( C(n, r) = \\frac{n!}{r!(n-r)!} \\), where \\( n \\) is the total number of individuals, and \\( r \\) is the number of individuals to choose.\n\nFor our scenario:\n\\[\nC(6, 3) = \\frac{6!}{3!(6-3)!} = \\frac{6 \\times 5 \\times 4}{3 \\times 2 \\times 1} = 20\n\\]\n\n### Step 3: Consider the Constraint\nNow we need to exclude the combinations where both Amy and Bob are included in the selection. To do this, we will calculate how many combinations include both A and B.\n\n### Step 4: Count Combinations Including Both A and B\nIf both A and B are selected, we need to choose 1 more individual from the remaining individuals {C, D, E, F} (4 individuals remaining).\n\nWe can compute the number of ways to choose 1 individual from 4:\n\\[\nC(4, 1) = 4\n\\]\nThe possible combinations that include both A and B are:\n1. A, B, C\n2. A, B, D\n3. A, B, E\n4. A, B, F\n\n### Step 5: Subtract Invalid Combinations from Total\nNow, we reduce the total combinations of 20 by the 4 combinations that include both A and B:\n\\[\n\\text{Valid combinations} = \\text{Total combinations} - \\text{Combinations including A and B}\n\\]\n\\[\n\\text{Valid combinations} = 20 - 4 = 16\n\\]\n\n### Final Step: List the Valid Combinations\nFor clarity, let's list the valid combinations that meet the criteria of not having both A and B together:\n\n1. A, C, D\n2. A, C, E\n3. A, C, F\n4. A, D, E\n5. A, D, F\n6. A, E, F\n7. B, C, D\n8. B, C, E\n9. B, C, F\n10. B, D, E\n11. B, D, F\n12. B, E, F\n13. C, D, E\n14. C, D, F\n15. C, E, F\n16. D, E, F\n\n### Conclusion\nThus, the total number of valid combinations, ensuring Amy and Bob are not serving together, is **16**.", 'To solve the problem of selecting a committee of 3 members from a group of 6 individuals, with the condition that Amy and Bob cannot both be included, we can break down the solution into steps.\n\n**Step 1: Count total combinations without restrictions.**\nWe first calculate the total number of ways to choose 3 members from the 6 individuals without any restrictions. We can use the binomial coefficient formula:\n\n\\[\n\\binom{n}{k} = \\frac{n!}{k!(n-k)!}\n\\]\n\nIn this case, \\( n = 6 \\) and \\( k = 3 \\):\n\n\\[\n\\binom{6}{3} = \\frac{6!}{3!(6-3)!} = \\frac{6 \\times 5 \\times 4}{3 \\times 2 \\times 1} = 20\n\\]\n\nSo, there are 20 ways to select any 3 individuals from 6.\n\n**Step 2: Count combinations including both Amy and Bob.**\nNext, we find the number of combinations where both Amy and Bob are included in the committee. If Amy and Bob are both included, we need to choose 1 more member from the remaining 4 individuals (since there are originally 6 individuals and 2 spots filled by Amy and Bob).\n\nThe number of ways to choose 1 member from the remaining 4 is:\n\n\\[\n\\binom{4}{1} = 4\n\\]\n\nThus, there are 4 combinations that include both Amy and Bob.\n\n**Step 3: Calculate valid combinations.**\nTo find the number of suitable combinations that meet the requirement of not having both Amy and Bob on the committee, we take the total combinations and subtract the combinations that include both Amy and Bob:\n\n\\[\n\\text{Valid combinations} = \\text{Total combinations} - \\text{Combinations with Amy and Bob}\n\\]\n\nSubstituting the values we found:\n\n\\[\n\\text{Valid combinations} = 20 - 4 = 16\n\\]\n\nTherefore, the number of suitable combinations of the committee that can be created under the given restrictions is:\n\n\\[\n\\boxed{16}\n\\]', "To solve the problem of forming a committee of 3 people from a pool of 6 people where Amy and Bob cannot be on the committee together, we can break down the solution into several steps:\n\n1. **Identify the members of the pool**:\n   - Let's label the 6 people as A (Amy), B (Bob), C, D, E, F (the others).\n\n2. **Calculate the total number of combinations without restrictions**:\n   - The total number of ways to choose 3 people from 6 is given by the combination formula:\n     \\[\n     \\binom{n}{r} = \\frac{n!}{r!(n-r)!}\n     \\]\n   - Here, \\( n = 6 \\) and \\( r = 3 \\):\n     \\[\n     \\binom{6}{3} = \\frac{6!}{3!3!} = \\frac{6 \\times 5 \\times 4}{3 \\times 2 \\times 1} = 20\n     \\]\n   - So, there are 20 ways to choose 3 people from 6 without any restrictions.\n\n3. **Account for the restriction (Amy and Bob can't be together)**:\n   - We need to calculate the number of combinations where both Amy and Bob are included, and then subtract this from the total.\n\n4. **Calculate the combinations where both Amy and Bob are on the committee**:\n   - If both Amy and Bob are on the committee, we only need to choose 1 more member from the remaining pool of 4 people (C, D, E, F).\n   - The number of ways to choose 1 additional member from 4 is:\n     \\[\n     \\binom{4}{1} = 4\n     \\]\n   - There are 4 ways to form a committee including both Amy and Bob.\n\n5. **Subtract the restricted combinations from the total**:\n   - The total valid combinations will therefore be:\n     \\[\n     \\text{Total combinations} - \\text{Combinations with Amy and Bob together} = 20 - 4 = 16\n     \\]\n\n6. **Final answer**:\n   - Thus, the number of viable combinations for the committee of 3 people, without both Amy and Bob being included together, is **16**.\n\nIn conclusion, the number of viable combinations is **16**."]

# eval_responses = await asyncio.gather(*eval_tasks)
# > [ResponseDetails(solution_number=16, correctness_probability=0.95), ResponseDetails(solution_number=16, correctness_probability=0.95), ResponseDetails(solution_number=16, correctness_probability=0.95)]

# response_scores
# {16: 2.8499999999999996}



import asyncio
import random

from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


@openai.call(model="gpt-4o-mini", call_params={"temperature": random.uniform(0, 2)})
@prompt_template(
    "Answer the following question, going through it step by step: {query}"
)
async def zero_shot_cot(query: str): ...


class ProblemInfo(BaseModel):
    key_info: list[str] = Field(
        ...,
        description="""Key pieces of information needed to solve the problem.
        Each key info is written as a short phrase.""",
    )


@openai.call(model="gpt-4o-mini", response_model=ProblemInfo)
@prompt_template(
    """
    For the following question, identify the key pieces of information
    needed to answer the question.
    Question: {query}
    """
)
def identify_key_info(query: str): ...


class MaskedPrompts(BaseModel):
    masked_prompts: list[str] = Field(
        ...,
        description="""The original prompt with the specified key info replaced with X.
        Example:
        Prompt: Jane walked 6 miles today. It was a sunny day.
        Key info: How many miles Jane walked
        Masked Prompt: Jane walked X miles today. It was a sunny day.""",
    )


@openai.call(model="gpt-4o-mini", response_model=MaskedPrompts)
@prompt_template(
    """
    Generate one masked copy of the prompt for each piece of key info.
    The number of masked prompts you generate should be equal to the number of pieces
    of key info, and each masked prompt should mask one key piece of information.
    Prompt: {query}
    Key Info: {key_info}
    """
)
async def mask_prompt(query: str) -> openai.OpenAIDynamicConfig:
    key_info = identify_key_info(query).key_info
    return {"computed_fields": {"key_info": key_info}}


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    SYSTEM:
    You will be given a masked prompt with some value Xd out, and the solution to
    the original problem. Return the prompt verbatim, but fill in the value for X
    according to what you see in the solution.
    USER:
    solution: {solution}
    masked prompt: {masked_prompt}
    """
)
async def fill_in_value(solution: str, masked_prompt: str): ...


class PromptComparison(BaseModel):
    score: int = Field(
        ...,
        description="""The number of variation prompts which are
                    semantically equivalent to the original prompt.
                    For numbers in the prompts, expect the same values, and for
                    quantitative words, only count semantically identical terms, such
                    as two times = double.""",
    )


@openai.call(model="gpt-4o-mini", response_model=PromptComparison)
@prompt_template(
    """
    SYSTEM:
    You will be given an original prompt and some variations of it. Return the
    number of variations which are semantically identical.
    USER:
    Original Prompt:
    {query}
    Variations:
    {numbered_variations}
    """
)
def score_variations(
    query: str, filled_in_prompts: list[str]
) -> openai.OpenAIDynamicConfig:
    return {
        "computed_fields": {
            "numbered_variations": [
                f"{i+1}. {_}" for i, _ in enumerate(filled_in_prompts)
            ]
        }
    }


async def evaluate_solution(query: str, solution: str, masked_prompts: list[str]):
    tasks = [fill_in_value(solution, masked_prompt) for masked_prompt in masked_prompts]
    responses = await asyncio.gather(*tasks)
    filled_in_prompts = [response.content for response in responses]
    score = score_variations(query, filled_in_prompts).score
    return score


query = """Tim wanted to make lemonade for a pool party. For a gallon of lemonade,
his recipe called for 1 cup of fresh lemon juice. He found that 6 lemons would yield
1 cup of juice. He figured he would need to make 4 gallons of lemonade for the
party. His best friend Allen asked if Tim could make an extra gallon for him that
was twice as tart as the other gallons. How many lemons will Tim need?"""


async def self_verify(query: str, num_solutions: int):
    tasks = [mask_prompt(query)]
    tasks += [zero_shot_cot(query) for _ in range(num_solutions)]
    responses = await asyncio.gather(*tasks)
    masked_prompts = responses[0].masked_prompts  # type: ignore
    # Uncomment to see intermediate responses
    print(masked_prompts)
    cot_responses = [response.content for response in responses[1:]]  # type: ignore
    print(cot_responses)
    tasks = [
        evaluate_solution(query, solution, masked_prompts) for solution in cot_responses
    ]
    scores = await asyncio.gather(*tasks)
    print(scores)
    print(f"MAX SCORE:{len(masked_prompts)}")
    for cot_solution, score in zip(cot_responses, scores, strict=False):
        print(f"Solution:\n{cot_solution}\nScore:\n{score}\n")


asyncio.run(self_verify(query, 3))
# MAX SCORE:4
# Solution:
# Let's break down the problem step by step.

# 1. **Determine the total amount of lemonade needed:**
#    - Tim originally planned to make 4 gallons of lemonade.
#    - Allen asked for an extra gallon, making the total amount of lemonade:
#      \[
#      4 \text{ gallons} + 1 \text{ gallon} = 5 \text{ gallons}
#      \]

# 2. **Calculate the amount of lemon juice needed for 5 gallons:**
#    - Each gallon requires 1 cup of lemon juice.
#    - Therefore, for 5 gallons, the total amount of lemon juice needed is:
#      \[
#      5 \text{ gallons} \times 1 \text{ cup/gallon} = 5 \text{ cups of lemon juice}
#      \]

# 3. **Determine the amount of lemon juice needed for the extra tart gallon:**
#    - Allen's gallon is twice as tart as the others. This means it will require double the amount of lemon juice.
#    - Therefore, for the extra gallon, the amount of lemon juice needed is:
#      \[
#      1 \text{ cup} \times 2 = 2 \text{ cups of lemon juice}
#      \]

# 4. **Calculate the total amount of lemon juice needed:**
#    - For the 4 gallons, Tim needs 4 cups of lemon juice (1 cup each).
#    - For the extra tart gallon, he needs 2 cups of lemon juice.
#    - Thus, the total amount of lemon juice needed is:
#      \[
#      4 \text{ cups} + 2 \text{ cups} = 6 \text{ cups of lemon juice}
#      \]

# 5. **Determine how many lemons are needed to make 6 cups of lemon juice:**
#    - From the information given, 6 lemons yield 1 cup of lemon juice.
#    - To find out how many lemons are needed for 6 cups, we calculate:
#      \[
#      6 \text{ lemons/cup} \times 6 \text{ cups} = 36 \text{ lemons}
#      \]

# So, Tim will need a total of **36 lemons** to make the lemonade for the pool party, including the extra tart gallon for Allen.
# Score:
# 0

# Solution:
# Let's go through the problem step by step.

# 1. **Determine the amount of lemon juice needed for 4 gallons of lemonade.**
#    - According to Tim's recipe, 1 gallon of lemonade requires 1 cup of lemon juice.
#    - For 4 gallons, the total amount of lemon juice needed is:
#      \[
#      4 \text{ gallons} \times 1 \text{ cup/gallon} = 4 \text{ cups of lemon juice}
#      \]

# 2. **Calculate the number of lemons needed for 4 gallons of lemonade.**
#    - Tim found that 6 lemons yield 1 cup of juice.
#    - To find out how many lemons are needed for 4 cups of juice, we first determine how many cups of juice can be obtained from 6 lemons:
#      \[
#      6 \text{ lemons} \rightarrow 1 \text{ cup}
#      \]
#    - Therefore, for 4 cups of juice, the number of lemons required is:
#      \[
#      4 \text{ cups} \times 6 \text{ lemons/cup} = 24 \text{ lemons}
#      \]

# 3. **Determine the amount of lemon juice needed for the extra gallon that is twice as tart.**
#    - Since the extra gallon is supposed to be twice as tart, it will require twice the amount of lemon juice as a regular gallon.
#    - Therefore, for the extra gallon:
#      \[
#      1 \text{ gallon} \times 2 \text{ cups/gallon} = 2 \text{ cups of lemon juice}
#      \]

# 4. **Calculate the number of lemons needed for the extra gallon.**
#    - Using the same conversion as before (6 lemons yield 1 cup of juice), we find out how many lemons are needed for 2 cups of juice:
#      \[
#      2 \text{ cups} \times 6 \text{ lemons/cup} = 12 \text{ lemons}
#      \]

# 5. **Calculate the total number of lemons needed.**
#    - Now, we can find the total number of lemons needed for both the 4 gallons of lemonade and the extra gallon:
#      \[
#      24 \text{ lemons (for 4 gallons)} + 12 \text{ lemons (for extra gallon)} = 36 \text{ lemons}
#      \]

# Thus, Tim will need a total of **36 lemons** to make the lemonade for the pool party and the extra tart gallon for Allen.
# Score:
# 4

# Solution:
# Let's break this down step by step to determine how many lemons Tim will need.

# ### Step 1: Determine the amount of lemon juice needed for 4 gallons of lemonade.

# The recipe calls for **1 cup of fresh lemon juice** for **1 gallon of lemonade**. Therefore, for **4 gallons**, Tim will need:

# \[
# 4 \text{ gallons} \times 1 \text{ cup/gallon} = 4 \text{ cups of lemon juice}
# \]

# ### Step 2: Calculate the lemon juice needed for Allen's extra gallon.

# Allen wants an extra gallon that is **twice as tart**. To make it twice as tart, Tim will need **2 cups of lemon juice** for that gallon (since the standard recipe uses 1 cup for 1 gallon).

# ### Step 3: Find the total amount of lemon juice needed.

# Now, we add the lemon juice needed for both Tim's and Allen's lemonade:

# \[
# 4 \text{ cups (for 4 gallons)} + 2 \text{ cups (for Allen's extra gallon)} = 6 \text{ cups of lemon juice}
# \]

# ### Step 4: Determine how many lemons are needed to yield 6 cups of juice.

# We know that **6 lemons yield 1 cup of juice**. To find out how many lemons are needed for **6 cups of juice**, we can set up the following calculation:

# \[
# \text{Number of lemons} = 6 \text{ lemons/cup} \times 6 \text{ cups} = 36 \text{ lemons}
# \]

# ### Conclusion

# Tim will need **36 lemons** to make 4 gallons of regular lemonade and 1 extra gallon that is twice as tart.
# Score:
# 0


# Intermediate responses

# masked_prompts = responses[0].masked_prompts
# > ['Tim wanted to make lemonade for a pool party. For X gallon of lemonade, his recipe called for 1 cup of fresh lemon juice. He found that 6 lemons would yield 1 cup of juice. He figured he would need to make 4 gallons of lemonade for the party. His best friend Allen asked if Tim could make an extra gallon for him that was twice as tart as the other gallons. How many lemons will Tim need?', 'Tim wanted to make lemonade for a pool party. For a gallon of lemonade, his recipe called for X cup of fresh lemon juice. He found that 6 lemons would yield 1 cup of juice. He figured he would need to make 4 gallons of lemonade for the party. His best friend Allen asked if Tim could make an extra gallon for him that was twice as tart as the other gallons. How many lemons will Tim need?', 'Tim wanted to make lemonade for a pool party. For a gallon of lemonade, his recipe called for 1 cup of fresh lemon juice. He found that X lemons would yield 1 cup of juice. He figured he would need to make 4 gallons of lemonade for the party. His best friend Allen asked if Tim could make an extra gallon for him that was twice as tart as the other gallons. How many lemons will Tim need?', 'Tim wanted to make lemonade for a pool party. For a gallon of lemonade, his recipe called for 1 cup of fresh lemon juice. He found that 6 lemons would yield 1 cup of juice. He figured he would need to make 4 gallons of lemonade for the party. His best friend Allen asked if Tim could make an extra X gallon for him that was twice as tart as the other gallons. How many lemons will Tim need?']

# print(cot_responses)
# > ["Let's break down the problem step by step.\n\n1. **Determine the total amount of lemonade needed:**\n   - Tim originally planned to make 4 gallons of lemonade.\n   - Allen asked for an extra gallon, making the total amount of lemonade:\n     \\[\n     4 \\text{ gallons} + 1 \\text{ gallon} = 5 \\text{ gallons}\n     \\]\n\n2. **Calculate the amount of lemon juice needed for 5 gallons:**\n   - Each gallon requires 1 cup of lemon juice.\n   - Therefore, for 5 gallons, the total amount of lemon juice needed is:\n     \\[\n     5 \\text{ gallons} \\times 1 \\text{ cup/gallon} = 5 \\text{ cups of lemon juice}\n     \\]\n\n3. **Determine the amount of lemon juice needed for the extra tart gallon:**\n   - Allen's gallon is twice as tart as the others. This means it will require double the amount of lemon juice.\n   - Therefore, for the extra gallon, the amount of lemon juice needed is:\n     \\[\n     1 \\text{ cup} \\times 2 = 2 \\text{ cups of lemon juice}\n     \\]\n\n4. **Calculate the total amount of lemon juice needed:**\n   - For the 4 gallons, Tim needs 4 cups of lemon juice (1 cup each).\n   - For the extra tart gallon, he needs 2 cups of lemon juice.\n   - Thus, the total amount of lemon juice needed is:\n     \\[\n     4 \\text{ cups} + 2 \\text{ cups} = 6 \\text{ cups of lemon juice}\n     \\]\n\n5. **Determine how many lemons are needed to make 6 cups of lemon juice:**\n   - From the information given, 6 lemons yield 1 cup of lemon juice.\n   - To find out how many lemons are needed for 6 cups, we calculate:\n     \\[\n     6 \\text{ lemons/cup} \\times 6 \\text{ cups} = 36 \\text{ lemons}\n     \\]\n\nSo, Tim will need a total of **36 lemons** to make the lemonade for the pool party, including the extra tart gallon for Allen.", "Let's go through the problem step by step.\n\n1. **Determine the amount of lemon juice needed for 4 gallons of lemonade.**\n   - According to Tim's recipe, 1 gallon of lemonade requires 1 cup of lemon juice.\n   - For 4 gallons, the total amount of lemon juice needed is:\n     \\[\n     4 \\text{ gallons} \\times 1 \\text{ cup/gallon} = 4 \\text{ cups of lemon juice}\n     \\]\n\n2. **Calculate the number of lemons needed for 4 gallons of lemonade.**\n   - Tim found that 6 lemons yield 1 cup of juice.\n   - To find out how many lemons are needed for 4 cups of juice, we first determine how many cups of juice can be obtained from 6 lemons:\n     \\[\n     6 \\text{ lemons} \\rightarrow 1 \\text{ cup}\n     \\]\n   - Therefore, for 4 cups of juice, the number of lemons required is:\n     \\[\n     4 \\text{ cups} \\times 6 \\text{ lemons/cup} = 24 \\text{ lemons}\n     \\]\n\n3. **Determine the amount of lemon juice needed for the extra gallon that is twice as tart.**\n   - Since the extra gallon is supposed to be twice as tart, it will require twice the amount of lemon juice as a regular gallon.\n   - Therefore, for the extra gallon:\n     \\[\n     1 \\text{ gallon} \\times 2 \\text{ cups/gallon} = 2 \\text{ cups of lemon juice}\n     \\]\n\n4. **Calculate the number of lemons needed for the extra gallon.**\n   - Using the same conversion as before (6 lemons yield 1 cup of juice), we find out how many lemons are needed for 2 cups of juice:\n     \\[\n     2 \\text{ cups} \\times 6 \\text{ lemons/cup} = 12 \\text{ lemons}\n     \\]\n\n5. **Calculate the total number of lemons needed.**\n   - Now, we can find the total number of lemons needed for both the 4 gallons of lemonade and the extra gallon:\n     \\[\n     24 \\text{ lemons (for 4 gallons)} + 12 \\text{ lemons (for extra gallon)} = 36 \\text{ lemons}\n     \\]\n\nThus, Tim will need a total of **36 lemons** to make the lemonade for the pool party and the extra tart gallon for Allen.", "Let's break this down step by step to determine how many lemons Tim will need.\n\n### Step 1: Determine the amount of lemon juice needed for 4 gallons of lemonade.\n\nThe recipe calls for **1 cup of fresh lemon juice** for **1 gallon of lemonade**. Therefore, for **4 gallons**, Tim will need:\n\n\\[\n4 \\text{ gallons} \\times 1 \\text{ cup/gallon} = 4 \\text{ cups of lemon juice}\n\\]\n\n### Step 2: Calculate the lemon juice needed for Allen's extra gallon.\n\nAllen wants an extra gallon that is **twice as tart**. To make it twice as tart, Tim will need **2 cups of lemon juice** for that gallon (since the standard recipe uses 1 cup for 1 gallon).\n\n### Step 3: Find the total amount of lemon juice needed.\n\nNow, we add the lemon juice needed for both Tim's and Allen's lemonade:\n\n\\[\n4 \\text{ cups (for 4 gallons)} + 2 \\text{ cups (for Allen's extra gallon)} = 6 \\text{ cups of lemon juice}\n\\]\n\n### Step 4: Determine how many lemons are needed to yield 6 cups of juice.\n\nWe know that **6 lemons yield 1 cup of juice**. To find out how many lemons are needed for **6 cups of juice**, we can set up the following calculation:\n\n\\[\n\\text{Number of lemons} = 6 \\text{ lemons/cup} \\times 6 \\text{ cups} = 36 \\text{ lemons}\n\\]\n\n### Conclusion\n\nTim will need **36 lemons** to make 4 gallons of regular lemonade and 1 extra gallon that is twice as tart."]

# scores
# > [0, 4, 0]



"""This module contains the `AzureCallResponseChunk` class.

usage docs: learn/streams.md#handling-streamed-responses
"""

from azure.ai.inference.models import (
    CompletionsFinishReason,
    CompletionsUsage,
    StreamingChatCompletionsUpdate,
)
from pydantic import SkipValidation

from ..base import BaseCallResponseChunk


class AzureCallResponseChunk(
    BaseCallResponseChunk[StreamingChatCompletionsUpdate, CompletionsFinishReason]
):
    """A convenience wrapper around the Azure `ChatCompletionChunk` streamed chunks.

    When calling the Azure API using a function decorated with `azure_call` and
    `stream` set to `True`, the stream will contain `AzureResponseChunk` instances with
    properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.azure import azure_call


    @azure_call("gpt-4o-mini", stream=True)
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    stream = recommend_book("fantasy")  # response is an `AzureStream`
    for chunk, _ in stream:
        print(chunk.content, end="", flush=True)
    ```
    """

    chunk: SkipValidation[StreamingChatCompletionsUpdate]

    @property
    def content(self) -> str:
        """Returns the content for the 0th choice delta."""
        delta = None
        if self.chunk.choices:
            delta = self.chunk.choices[0].delta
        return delta.content if delta is not None and delta.content else ""

    @property
    def finish_reasons(self) -> list[CompletionsFinishReason]:
        """Returns the finish reasons of the response."""
        return [
            finish_reason
            if isinstance(finish_reason, CompletionsFinishReason)
            else CompletionsFinishReason(finish_reason)
            for choice in self.chunk.choices
            if (finish_reason := choice.finish_reason)
        ]

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.chunk.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.chunk.id

    @property
    def usage(self) -> CompletionsUsage:
        """Returns the usage of the chat completion."""
        return self.chunk.usage

    @property
    def input_tokens(self) -> int:
        """Returns the number of input tokens."""
        return self.usage.prompt_tokens

    @property
    def output_tokens(self) -> int:
        """Returns the number of output tokens."""
        return self.usage.completion_tokens



from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


class Translations(BaseModel):
    translations: list[str] = Field(
        ..., description="The list of translations into the requested language"
    )


@openai.call(model="gpt-4o-mini", response_model=Translations)
@prompt_template(
    """
    For this phrase: {phrase}


    Give {num_translations} translations in {language}
    """
)
def translate(phrase: str, language: str, num_translations: int): ...


def prompt_paraphrasing(query: str, num_translations: int = 3):
    spanish_translations = translate(
        phrase=query,
        language="Spanish",
        num_translations=num_translations,
    )
    # Uncomment to see intermediate responses
    print(spanish_translations)
    # Avoid Duplicates
    prompt_varations = set()
    for spanish_phrase in spanish_translations.translations:
        back_translations = translate(
            spanish_phrase, language="English", num_translations=3
        )
        prompt_varations.update(back_translations.translations)
    return prompt_varations


print(
    prompt_paraphrasing(
        query="What are some manageable ways to improve my focus and productivity?"
    )
)
# > {'What are some manageable ways to improve my focus and productivity?', 'What manageable ways are there to improve my focus and productivity?', 'How can I improve my focus and productivity in a manageable way?', 'Which are some manageable ways to boost my focus and productivity?', 'What can I do to boost my focus and productivity in a manageable fashion?', 'What manageable methods are there to boost my focus and productivity?', 'Which manageable forms are there to enhance my focus and productivity?', 'How can I enhance my focus and productivity in a manageable manner?', 'What are some practical ways to enhance my focus and productivity?'}


# Intermediate Responses
# spanish translations
# > ['¿Cuáles son algunas maneras manejables de mejorar mi enfoque y productividad?', '¿Qué formas manejables hay para mejorar mi enfoque y productividad?', '¿Cómo puedo mejorar mi enfoque y productividad de manera manejable?']



from typing import Annotated

from pydantic import AfterValidator, BaseModel, Field, ValidationError

from mirascope.core import anthropic, prompt_template


class SpellingAndGrammarCheck(BaseModel):
    has_errors: bool = Field(
        description="Whether the text has typos or grammatical errors"
    )


@anthropic.call(
    model="claude-3-5-sonnet-20240620",
    response_model=SpellingAndGrammarCheck,
    json_mode=True,
)
@prompt_template(
    """
    Does the following text have any typos or grammatical errors? {text}
    """
)
def check_for_errors(text: str): ...


class TextWithoutErrors(BaseModel):
    text: Annotated[
        str,
        AfterValidator(
            lambda t: t
            if not (check_for_errors(t)).has_errors
            else (_ for _ in ()).throw(ValueError("Text contains errors"))
        ),
    ]


valid = TextWithoutErrors(text="This is a perfectly written sentence.")

try:
    invalid = TextWithoutErrors(
        text="I walkd to supermarket and i picked up some fish?"
    )
except ValidationError as e:
    print(f"Validation error: {e}")
    # > Validation error: 1 validation error for TextWithoutErrors
    #   text
    #     Value error, Text contains errors [type=value_error, input_value='I walkd to supermarket a... i picked up some fish?', input_type=str]
    #       For further information visit https://errors.pydantic.dev/2.8/v/value_error



import asyncio
from datetime import datetime

import aiohttp
import requests
from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel

from mirascope.core import openai, prompt_template

NIMBLE_TOKEN = "YOUR_NIMBLE_API_KEY"


class LocalizedRecommender(BaseModel):
    history: list[ChatCompletionMessageParam] = []

    async def _nimble_google_maps_places(
        self, session: aiohttp.ClientSession, place_id: str
    ):
        """
        Use Nimble to get the details of a place on Google Maps.
        """
        url = "https://api.webit.live/api/v1/realtime/serp"
        headers = {
            "Authorization": f"Basic {NIMBLE_TOKEN}",
            "Content-Type": "application/json",
        }
        place_data = {
            "parse": True,
            "search_engine": "google_maps_place",
            "place_id": place_id,
            "domain": "com",
            "format": "json",
            "render": True,
            "country": "US",
            "locale": "en",
        }
        async with session.get(url, json=place_data, headers=headers) as response:
            data = await response.json()
            result = data["parsing"]["entities"]["Place"][0]
            return {
                "opening_hours": result.get("place_information", {}).get(
                    "opening_hours", ""
                ),
                "rating": result.get("rating", ""),
                "name": result.get("title", ""),
            }

    async def _nimble_google_maps(self, latitude: float, longitude: float, query: str):
        """
        Use Nimble to search for places on Google Maps.
        """
        url = "https://api.webit.live/api/v1/realtime/serp"
        headers = {
            "Authorization": f"Basic {NIMBLE_TOKEN}",
            "Content-Type": "application/json",
        }
        search_data = {
            "parse": True,
            "search_engine": "google_maps_search",
            "query": query,
            "coordinates": {"latitude": latitude, "longitude": longitude},
            "domain": "com",
            "format": "json",
            "render": True,
            "country": "US",
            "locale": "en",
        }
        search_response = requests.post(url, headers=headers, json=search_data)
        search_json_response = search_response.json()
        search_results = [
            {
                "place_id": result.get("place_id", ""),
            }
            for result in search_json_response["parsing"]["entities"]["SearchResult"]
        ]
        results = []
        async with aiohttp.ClientSession() as session:
            tasks = [
                self._nimble_google_maps_places(session, results.get("place_id", ""))
                for results in search_results
            ]
            results = await asyncio.gather(*tasks)
        return results

    async def _get_current_date(self):
        """Get the current date and time."""
        return datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    async def _get_coordinates_from_location(self, location_name: str):
        """Get the coordinates of a location."""
        base_url = "https://nominatim.openstreetmap.org/search"
        params = {"q": location_name, "format": "json", "limit": 1}
        headers = {"User-Agent": "mirascope/1.0"}
        response = requests.get(base_url, params=params, headers=headers)
        data = response.json()

        if data:
            latitude = data[0].get("lat")
            longitude = data[0].get("lon")
            return f"Latitude: {latitude}, Longitude: {longitude}"
        else:
            return "No location found, ask me about a specific location."

    @openai.call(model="gpt-4o", stream=True)
    @prompt_template(
        """
        SYSTEM:
        You are a local guide that recommends the best places to visit in a place.
        Use the `_get_current_date` function to get the current date.
        Use the `_get_coordinates_from_location` function to get the coordinates of a location if you need it.
        Use the `_nimble_google_maps` function to get the best places to visit in a location based on the users query.

        MESSAGES: {self.history}
        USER: {question}
        """
    )
    async def _step(self, question: str) -> openai.OpenAIDynamicConfig:
        return {
            "tools": [
                self._get_current_date,
                self._get_coordinates_from_location,
                self._nimble_google_maps,
            ]
        }

    async def _get_response(self, question: str):
        response = await self._step(question)
        tool_call = None
        output = None
        async for chunk, tool in response:
            if tool:
                output = await tool.call()
                tool_call = tool
            else:
                print(chunk.content, end="", flush=True)
        if response.user_message_param:
            self.history.append(response.user_message_param)
        self.history.append(response.message_param)
        if tool_call and output:
            self.history += response.tool_message_params([(tool_call, str(output))])
            return await self._get_response(question)
        return

    async def run(self):
        while True:
            question = input("(User): ")
            if question == "exit":
                break
            print("(Assistant): ", end="", flush=True)
            await self._get_response(question)
            print()


recommender = LocalizedRecommender(history=[])
asyncio.run(recommender.run())



from mirascope.core import anthropic, mistral, openai, prompt_template


@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


openai_recommendation = openai.call("gpt-4o-mini")(recommend_book)("fantasy")
print("OpenAI recommendation:", openai_recommendation.content)

anthropic_recommendation = anthropic.call("claude-3-5-sonnet-20240620")(recommend_book)(
    "fantasy"
)
print("Anthropic recommendation:", anthropic_recommendation.content)

mistral_recommendation = mistral.call("mistral-large-latest")(recommend_book)("fantasy")
print("Mistral recommendation:", mistral_recommendation.content)



"""The `gemini_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import GeminiCallParams
from .call_response import GeminiCallResponse
from .call_response_chunk import GeminiCallResponseChunk
from .dynamic_config import GeminiDynamicConfig
from .stream import GeminiStream
from .tool import GeminiTool

gemini_call = call_factory(
    TCallResponse=GeminiCallResponse,
    TCallResponseChunk=GeminiCallResponseChunk,
    TDynamicConfig=GeminiDynamicConfig,
    TStream=GeminiStream,
    TToolType=GeminiTool,
    TCallParams=GeminiCallParams,
    default_call_params=GeminiCallParams(),
    setup_call=setup_call,
    get_json_output=get_json_output,
    handle_stream=handle_stream,  # pyright: ignore [reportArgumentType]
    handle_stream_async=handle_stream_async,  # pyright: ignore [reportArgumentType]
)
"""A decorator for calling the Gemini API with a typed function.

usage docs: learn/calls.md

This decorator is used to wrap a typed function that calls the Gemini API. It parses
the prompt template of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.gemini import gemini_call


@gemini_call("gemini-1.5-flash")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str):
    ...

response = recommend_book("fantasy")
print(response.content)
```

Args:
    model (str): The Gemini model to use in the API call.
    stream (bool): Whether to stream the response from the API call.
    tools (list[BaseTool | Callable]): The tools to use in the Gemini API call.
    response_model (BaseModel | BaseType): The response model into which the response
        should be structured.
    output_parser (Callable[[GeminiCallResponse | ResponseModelT], Any]): A function
        for parsing the call response whose value will be returned in place of the
        original call response.
    json_modem (bool): Whether to use JSON Mode.
    client (object): An optional custom client to use in place of the default client.
    call_params (GeminiCallParams): The `GeminiCallParams` call parameters to use in the
        API call.

Returns:
    decorator (Callable): The decorator for turning a typed function into a Gemini API
        call.
"""
